// Video stream blink detection - Stdin version

// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "blinkdetect"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#include "common.h"
#include "configfile.h"
#include "mqtt.h"
#include "stdin.h"
#include "process_image.h"
#include "../shared/version.h"
#include "../shared/error.h"

const char *progname = PROGNAME;




void
print_help (FILE * f)
{
  fprintf (f, "\n"
	   "USAGE: ..videostream.. | %s [options]\n"
	   "\n"
	   "SCWIPS Video stream blink detection\n"
	   "\n"
	   "Options are:\n"
	   "    --cfg <filename>   Configuration file name\n"
	   "\n"
	   "    --help             Print this help\n"
	   "\n"
	   "Reconnections to mqtt broker are handled gracefully.\n"
	   "DNS and connect are protected by alarm(), use external loop to restart.\n"
	   "\nPart of SCWIPS version %s\n\n", PROGNAME, SCWIPSVERSION);
}


/* Parse commandline parameters.
   Returns:
   0 = all okay, continue
   1 = immediately exit successfully (after --help)
   2 = immediately exit unsuccessfully
*/
int
parse_cmdline (int argc, char **argv, scwstate_t * state)
{
  int i;

  // Rather simplistic & easy parameter parsing, adapted from earlier stuff
  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "--help"))
	{
	  print_help (stdout);
	  return 1;		// exit successfully
	}

      else if (!strcmp (argv[i], "--cfg"))
	{
	  i++;
	  if (i >= argc)
	    CMDLINEERR_RETURN2 ("parameter --cfg requires argument");
	  state->configfilename = argv[i];
	}

      else
	{
	  CMDLINEERR_RETURN2 ("unknown parameter: %s", argv[i]);
	}
    }

  if (state->configfilename == NULL || state->configfilename[0] == '\0')
    CMDLINEERR_RETURN2 ("please specify --cfg configfilename");

  // All should be okay now
  return 0;
}


int
main (int argc, char **argv)
{
  scwstate_t realstate;
  scwstate_t *state;
  int res;

  alarm (5);

  state = &realstate;		// Easy pointer-everywhere
  memset (state, 0, sizeof (*state));

  res = parse_cmdline (argc, argv, state);
  if (res >= 1)
    exit (res - 1);

  res = parse_configfile (state);
  if (res != 0)
    exit (1);

  res = mqtt_init (state);
  if (res != 0)
    exit (1);

  stdin_init (state);

  while (1)
    {
      res = mqtt_selectround (state, 900000, 1);
      if (res < 0)
	// Neat shutdown or grave error
	break;
      if (res == 1)
	{
	  res = read_stdin (state, 0);
	  if (res < 0)
	    // Some grave error
	    break;
	  if (res == 1)
	    {
	      //MDBG ("img received");
	      process_stdin_image (state);
	      alarm (5);
	    }
	}
    }

  mqtt_stop (state);

  return 0;
}
