#ifndef _BLINKDETECT_COMMON_H
#define _BLINKDETECT_COMMON_H

#include <time.h>		// for time_t
#include <unistd.h>		// for getpid
#include <jansson.h>
#include "../shared/mqttpub.h"	// for TOPICLEN [alt: local var, then strdup]


#define HISTIMGS		3	/* off - half - on, or vice versa */


#define MDBG(format, args...)  \
    mqttprintf(state->mosq,  \
	state->topicprefix, "cam", state->camname, "debug", NULL,  \
	"%s %d %s:%d %s(): " format ,  \
	progname, getpid(), __FILE__, __LINE__, __func__, ##args)

#define noMDBG(args...)		// nothing: for easy commenting-out


// Values for recvimg_formatbase and -detail
#define IMGF_NONE	0
#define IMGF_RGB	1
#define IMGF_YUV	2
#define IMGF_410p	3	// detail for IMGF_YUV
#define IMGF_411p	4
#define IMGF_420p	5
#define IMGF_422p	6
#define IMGF_440p	7
#define IMGF_444p	8

// Values for detect_state
#define DSTATE_WAIT		0
#define DSTATE_FIRSTRISE	1
#define DSTATE_IGNORERISE	2
#define DSTATE_FIRSTFALL	3
#define DSTATE_IGNOREFALL	4


// From ffmpeg libavutil/common.h
// E.g. shift 2 bits: first add 3 (=4-1)
#define CEIL_RSHIFT(a,b)  ( ((a) + (1<<(b)) - 1) >> (b) )


typedef struct
{
  int blinkw;
  int blinkh;
  double blinkx_d;
  double blinky_d;
  long weightsum;
  int brightestblink;
  int brightestx;
  int brightesty;
} scwedge_t;

typedef struct
{
  int *img;			// colorfiltered image, malloc'd
  int area_xmin;		// available colorfiltered area (min .. max-1)
  int area_xmax;
  int area_ymin;
  int area_ymax;

  scwedge_t rise;
  scwedge_t fall;
} scwhistimg_t;


typedef struct
{
  char *configfilename;

  // mqtt params

  struct mosquitto *mosq;
  const char *svr_hostname;
  int svr_port;
  int connstate;
  time_t connect_nexttime;	/* Planned time of next connect attempt;
				   also stdin-EOF force-disconnect time */
  const char *topicprefix;	// strdup = malloc'd

  char broadcmd_topic[TOPICLEN + 5];
  char allcamcmd_topic[TOPICLEN + 5];
  char camcmd_topic[TOPICLEN + 5];
  char calcpix_topic[TOPICLEN + 5];

  const char *shutdown_touchfile;


  // camera params

  const char *camname;		// config settings, strdup'd = malloc'd
  json_t *jcamtype;
  json_t *jcamdistortion;


  // image params and storage

  int width_config;
  int height_config;
  int width;
  int height;
  int maxdimension;		// max(width, height) for max margins

  scwhistimg_t histimg[HISTIMGS];	// color-filtered images history
  int curhist;			// round-robin index
  unsigned char *pubimg;	// image publish buffer, malloc'd
  int img_init_flag;

  int pubimg_orig_flag;
  int pubimg_colfilt_flag;


  // stdin state and storage

  unsigned char *headbuf;	// incoming header buffer when pipe-in, malloc'd
  int headbuf_size;		// large enough for any header
  int inheadbuf;
  int readinghead;

  unsigned char *recvimg;	// incoming image buffer when pipe-in, malloc'd
  int recvimg_size;		// *exact* image bytes
  int inrecvimg;
  int recvimg_formatbase;	// IMGF_* values
  int recvimg_formatdetail;	// IMGF_* values
  int recvimg_yuv_hshift;
  int recvimg_yuv_vshift;
  unsigned char *recvimg_yuv_plane[3];	// Starts of Y/Cb/Cr inside recvimg
  int recvimg_yuv_cbcrwidth;
  int recvimg_yuv_cbcrheight;


  // color filter params

  int colorfilter_r;		// config settings
  int colorfilter_g;
  int colorfilter_b;
  int colorfilter_cb_sethd;	// 0..255 value [16..240]
  int colorfilter_cr_sethd;
  int colorfilter_cb_autohd;
  int colorfilter_cr_autohd;
  int colorfilter_autobound_hd;
  int colorfilter_adjustreport;	// Repeating countdown timer


  // blink detection params & state

  int ourarea_x;		// area of interest, center + margins
  int ourarea_y;		// center moves _towards_ detected blink
  int ourarea_margin;
  int focusarea_min;		// config setting

  int backcalcarea_x;		// center _is_ backcalc pixel
  int backcalcarea_y;
  int backcalcarea_margin;

  int blinkthres_min;		// config setting
  int blinkthres_cur_hd;	// <<8; slowly adjusted to get blinkpixtarget
  int weightsum_slowfilt;	// <<8; detect sudden increase/decrease in pixels

  int *detectrise_img;		// helper storage for edge detection, malloc'd
  int *detectfall_img;

  int *weight_cols_rise;	// malloc'd
  int *weight_rows_rise;
  int *weight_cols_fall;
  int *weight_rows_fall;

  int detect_state;		// DSTATE_* values
  int detect_wait_since;	// frame counter
  int detect_wait_max;		// config setting: max #frames between edges

  int brightestx;
  int brightesty;


  // timeouts and counters

  int imgpersec_time;		// input fps timeout detection
  int imgpersec_cnt;
  int imgpersec_min;		// config setting

  int blinkspersec_cnt;
  int blinkspersec_time;

  int fixlost_timeout;		// blink timeout detection, counts frames

} scwstate_t;


#endif /* _BLINKDETECT_COMMON_H */
