// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#define PROGNAME  "ffmpeg-blinkdetect"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "../common.h"
#include "../configfile.h"
#include "../mqtt.h"
#include "../../shared/error.h"

const char *progname = PROGNAME;

static scwstate_t *global_blinkdetect_state = NULL;


// Returns:
// 1: okay, done.
// 0: not handled, caller should still sleep.
//
// Ffmpeg integration considers this a header and demands a prototype.
int blinkdetect_selectround_ffmpeg (int usec);
int
blinkdetect_selectround_ffmpeg (int usec)
{
  int res;

  // State is inside the blinkdetect muxer private data, can't normally
  // access it from main ffmpeg loop. So keep one static pointer,
  // should work as long as only one blinkdetect muxer is active.

  if (global_blinkdetect_state == NULL)
    return 0;

  res = mqtt_selectround (global_blinkdetect_state, usec, 0);
  if (res < 0)
    exit (1);

  return 1;
}


static int
blinkdetect_ffmpeg_init (scwstate_t * state)
{
  int res;

  alarm (5);

  memset (state, 0, sizeof (*state));

  global_blinkdetect_state = state;

  state->configfilename = getenv ("BLINKDETECT_CONF");
  if (state->configfilename == NULL || state->configfilename[0] == '\0')
    {
      STDERR ("Please set BLINKDETECT_CONF environment variable to "
	      "configuration file name");
      return 1;
    }

  res = parse_configfile (state);
  if (res != 0)
    return 1;

  res = mqtt_init (state);
  if (res != 0)
    return 1;

  return 0;
}
