// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

static int
blinkdetect_process_packet (AVFormatContext * s, AVPacket * pkt)
{
  AVStream *st = s->streams[pkt->stream_index];
  AVFrame *frame;
  int width, height, h_chroma_shift, v_chroma_shift;
  scwstate_t *state = s->priv_data;

  alarm (5);

  width = st->codecpar->width;
  height = st->codecpar->height;

  if (!state->img_init_flag)
    {
      state->width = width;
      state->height = height;
      image_init (state);
    }
  else
    {
      if (width != state->width || height != state->height)
	{
	  av_log (s, AV_LOG_ERROR,
		  "Mid-stream image size change is not supported.\n");
	  return AVERROR (EINVAL);
	}
    }

  av_pix_fmt_get_chroma_sub_sample (st->codecpar->format, &h_chroma_shift,
				    &v_chroma_shift);

  frame = (AVFrame *) pkt->data;

  // Remainder follows process_stdin_image()

  if (state->pubimg_orig_flag)
    {
      pubimg_planaryuv (state, "orig", state->width, state->height,
			h_chroma_shift, v_chroma_shift,
			frame->data, frame->linesize);
      state->pubimg_orig_flag = 0;
    }

  prepare_histslot (state);

  colfilt_planaryuv (state, h_chroma_shift, v_chroma_shift,
		     frame->data, frame->linesize);

  if (state->pubimg_colfilt_flag)
    {
      pubimg_intgrey (state, "colfilt", state->width, state->height,
		      state->histimg[state->curhist].img);
      state->pubimg_colfilt_flag = 0;
    }

  detect_blink (state);

  updates_afterimage (state);

  if (state->brightestx >= 0)
    {
      filteradjust_planaryuv (state, h_chroma_shift, v_chroma_shift,
			      frame->data, frame->linesize);
    }

  return 0;
}
