#ifndef _BLINKDETECT_IMAGE_INIT_H
#define _BLINKDETECT_IMAGE_INIT_H

#include "common.h"

static inline void
rgb2ycbcr (int r, int g, int b, int *y, int *cb, int *cr)
{
  // Formula: https://en.wikipedia.org/wiki/YCbCr  # ITU-R BT.601
  // r,g,b are 0..255 so won't go negative; but int so no problem anyway.

  *y = 16.49 + (65.738 * ((double) r)
		+ 129.057 * ((double) g) + 25.064 * ((double) b)) / 256.;

  *cb = 128.49 + (-37.945 * ((double) r)
		  - 74.494 * ((double) g) + 112.439 * ((double) b)) / 256.;

  *cr = 128.49 + (112.439 * ((double) r)
		  - 94.154 * ((double) g) - 18.285 * ((double) b)) / 256.;
}

static inline void
ycbcr2rgb (int y, int cb, int cr, int *r, int *g, int *b)
{
  // Formula: https://en.wikipedia.org/wiki/YCbCr  # ITU-R BT.601
  // May go beyond bounds a little; but int so no problem anyway.

  *r = -222.921 + 0.49 + (298.082 * ((double) y)
			  + 408.583 * ((double) cr)) / 256.;

  *g = 135.576 + 0.49 + (298.082 * ((double) y)
			 - 100.291 * ((double) cb)
			 - 208.120 * ((double) cr)) / 256.;

  *b = -276.836 + 0.49 + (298.082 * ((double) y)
			  + 516.412 * ((double) cb)) / 256.;
}


void image_init (scwstate_t * state);

#endif /* _BLINKDETECT_IMAGE_INIT_H */
