// Camera functions: various types of focus & distortion

#ifndef _CAMFUNCS_H
#define _CAMFUNCS_H

#include "../cmpfit/mpfit.h"


// Functions to map x,y,z point through camera's 0,0,0 focal point
// to x,y,"-f" on image "plane".
// I.e. functions must assume that camera is at origin, and positive-z
// is camera optical axis, and camera picture center is at x,y=0,0.
// Can take any number of params via array.

typedef void (camtype_func_t) (double objx, double objy, double objz,
			       double *camx, double *camy, double *params);
typedef void (camtype_initfunc_t) (double *param, mp_par * param_settings);


// Distortion functions to obtain real camera pixel from "officially correct"
// pixel. Can take any number of params via array.

typedef void (distortion_func_t) (double officialx, double officialy,
				  double *distortedx, double *distortedy,
				  double *params);
typedef void (distortion_initfunc_t) (double *param, mp_par * param_settings);


typedef struct
{
  char *name;
  int nparam;
  camtype_func_t *func;
  camtype_initfunc_t *initfunc;
} camtype_t;

typedef struct
{
  char *name;
  int nparam;
  distortion_func_t *func;
  distortion_initfunc_t *initfunc;
} camdistortion_t;

extern camtype_t camtype[];
extern camdistortion_t camdistortion[];


#endif /* _CAMFUNCS_H */
