// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: camtype_func_t camtype_basic;
// DECL: camtype_initfunc_t camtype_basic_init;
// REG_CAMTYPE: {"basic", 1, camtype_basic, camtype_basic_init}

void
camtype_basic (double objx, double objy, double objz, double *camx,
	       double *camy, double *params)
{
  // Focal distance, basic camera's have equal in x and y direction.
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  double f = params[0];
  double flip, newobjz;

  //*camx = f * objx / objz;
  //*camy = f * objy / objz;
  /* These are the well-known calculations, but they fail for objz <= 0.

     But actually everything with, e.g., objz < 0.001 will never be seen
     by the camera [NOTE: this is no fisheye!].

     We want to keep the formula linear in f and in objx/y to get the same
     derivatives.
     Also we want useful derivatives of objz that point in the right direction.
     These can be obtained if we somehow map objz [-inf...0.001] to [0...0.001].
     That's easy enough:
     -1/x maps [-inf...-1] to [0...+1], goes through (-1,1) with nice slope 1.
     Translate to (1,1) gives -1/(x-2), which maps [-inf...+1] to [0...+1].
     Then scalinmg x and y to (0.001,0,001) becomes -0.001/(x/0.001 - 2).

     Maple:
     newobjz := (objz, flip) -> \
     piecewise(objz < flip, -flip/((objz/flip) - 2), objz);
     plot(newobjz(objz, 0.1), objz=-0.3..0.3);
     plot(10/newobjz(objz, 0.1), objz=-1..1);
   */

  flip = 0.001;
  if (objz >= flip)
    newobjz = objz;
  else
    newobjz = -flip / ((objz / flip) - 2.);

  *camx = f * objx / newobjz;
  *camy = f * objy / newobjz;
}

void
camtype_basic_init (double *param, mp_par * param_settings)
{
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  param[0] = 1.;
  // We want positive focusdist (and not zero)
  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = 0.001;	// lower bound value
  param_settings[0].step = 0.001;
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 2.;	// advance limit value

  // Note: don't forget adjusting #params in REG_CAMTYPE above
}
