// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: camtype_func_t camtype_xyanglefocus;
// DECL: camtype_initfunc_t camtype_xyanglefocus_init;
// REG_CAMTYPE: {"xyanglefocus", 3, camtype_xyanglefocus, camtype_xyanglefocus_init}

void
camtype_xyanglefocus (double objx, double objy, double objz, double *camx,
		      double *camy, double *params)
{
  // Focal distance separately in one vs. other direction, any axis rotation.
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  double fx = params[0];
  double fy = params[1] * fx;	// == factor
  double theta = params[2];
  double flip, newobjz;
  double rotx, roty, focx, focy;

  // See newobjz mapping in camtype_basic() above.
  flip = 0.001;
  if (objz >= flip)
    newobjz = objz;
  else
    newobjz = -flip / ((objz / flip) - 2.);

  // First rotate x and y
  // http://planning.cs.uiuc.edu/node98.html
  rotx = objx * cos (theta) - objy * sin (theta);
  roty = objx * sin (theta) + objy * cos (theta);

  focx = fx * rotx / newobjz;
  focy = fy * roty / newobjz;

  // Then rotate back
  *camx = focx * cos (-theta) - focy * sin (-theta);
  *camy = focx * sin (-theta) + focy * cos (-theta);
}

void
camtype_xyanglefocus_init (double *param, mp_par * param_settings)
{
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  param[0] = 1.;
  param[1] = 1.2;		// factor!, start !=1 so that angle matters
  param[2] = 0.001;

  // We want positive focusdist (and not zero)
  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = 0.001;	// lower bound value
  param_settings[0].step = 0.001;
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 2.;	// advance limit value
  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = 0.001;	// lower bound value
  param_settings[1].step = 0.001;
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 2.;	// advance limit value

  // And angle is periodic.
  // Do not limit +/- 90' : solution will generally keep pushing to
  // the limit and not flip back to the other end.
  param_settings[2].periodic = 1;	// lower bound active
  param_settings[2].period = 2. * M_PI;	// wraps into -period/2 .. period/2
  param_settings[2].advlimited = 1;	// limit advance per iteration
  param_settings[2].advlimit = 2.;	// advance limit value
  param_settings[2].step = 0.001;

  // Note: don't forget adjusting #params in REG_CAMTYPE above
}
