// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: camtype_func_t camtype_xyfocus;
// DECL: camtype_initfunc_t camtype_xyfocus_init;
// REG_CAMTYPE: {"xyfocus", 2, camtype_xyfocus, camtype_xyfocus_init}

void
camtype_xyfocus (double objx, double objy, double objz, double *camx,
		 double *camy, double *params)
{
  // Focal distance separately in x vs. y direction.
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  double fx = params[0];
  double fy = params[1] * fx;	// == factor
  double flip, newobjz;

  // See newobjz mapping in camtype_basic() above.
  flip = 0.001;
  if (objz >= flip)
    newobjz = objz;
  else
    newobjz = -flip / ((objz / flip) - 2.);

  *camx = fx * objx / newobjz;
  *camy = fy * objy / newobjz;
}

void
camtype_xyfocus_init (double *param, mp_par * param_settings)
{
  // NOTE: param[0] is base focus, fixed to 1. during calib.
  param[0] = 1.;
  param[1] = 1.;
  // We want positive focusdist (and not zero)
  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = 0.001;	// lower bound value
  param_settings[0].step = 0.001;
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 2.;	// advance limit value
  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = 0.001;	// lower bound value
  param_settings[1].step = 0.001;
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 2.;	// advance limit value

  // Note: don't forget adjusting #params in REG_CAMTYPE above
}
