// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_024;
// DECL: distortion_initfunc_t distortion_024_init;
// REG_DISTORTION: {"024", 2, distortion_024, distortion_024_init}

// model: xd = x * ( 1 + k1 * r^2 + k2 * r^4 )

void
distortion_024 (double officialx, double officialy, double *distortedx,
		double *distortedy, double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are equal in x and y direction.
      // Distortion center is fixed at 0,0 in this model.
      double a = params[0];
      double b = params[1];
      double rsq = officialx * officialx + officialy * officialy;
      double fact;

      fact = (1 + a * rsq + b * rsq * rsq);

      *distortedx = officialx * fact;
      *distortedy = officialy * fact;
    }
  while (0);
}

void
distortion_024_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;

  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = -0.5;	// lower bound value
  param_settings[0].limited[1] = 1;	// upper bound active
  param_settings[0].limits[1] = 0.5;	// upper bound value
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 0.2;	// advance limit value

  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = -0.5;	// lower bound value
  param_settings[1].limited[1] = 1;	// upper bound active
  param_settings[1].limits[1] = 0.5;	// upper bound value
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
