// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_1div12;
// DECL: distortion_initfunc_t distortion_1div12_init;
// REG_DISTORTION: {"1div12", 3, distortion_1div12, distortion_1div12_init}

// model: xd = x * ( 1 + k1 * r ) / ( 1 + k2 * r + k3 * r^2 )
// https://arxiv.org/abs/cs/0307047v1

void
distortion_1div12 (double officialx, double officialy, double *distortedx,
		   double *distortedy, double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are equal in x and y direction.
      // Distortion center is fixed at 0,0 in this model.
      double a = params[0];
      double b = params[1];
      double c = params[2];
      double r = hypot (officialx, officialy);
      double fact;

      fact = (1 + a * r) / (1 + b * r + c * r * r);

      *distortedx = officialx * fact;
      *distortedy = officialy * fact;
    }
  while (0);
}

void
distortion_1div12_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
