// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_1div12shifted;
// DECL: distortion_initfunc_t distortion_1div12shifted_init;
// REG_DISTORTION: {"1div12shifted", 5, distortion_1div12shifted, distortion_1div12shifted_init}

void
distortion_1div12shifted (double officialx, double officialy,
			  double *distortedx, double *distortedy,
			  double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are equal in x and y direction.
      double a = params[0];
      double b = params[1];
      double c = params[2];
      // Distortion center is shifted by params.
      double centerx = params[3];
      double centery = params[4];

      officialx -= centerx;
      officialy -= centery;
      double r = hypot (officialx, officialy);
      double fact;

      fact = (1 + a * r) / (1 + b * r + c * r * r);

      *distortedx = officialx * fact;
      *distortedy = officialy * fact;
      *distortedx += centerx;
      *distortedy += centery;
    }
  while (0);
}

void
distortion_1div12shifted_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;
  param[3] = .001;
  param[4] = .001;

  param_settings[3].limited[0] = 1;	// lower bound active
  param_settings[3].limits[0] = -0.5;	// lower bound value
  param_settings[3].limited[1] = 1;	// upper bound active
  param_settings[3].limits[1] = 0.5;	// upper bound value
  param_settings[3].advlimited = 1;	// limit advance per iteration
  param_settings[3].advlimit = 0.2;	// advance limit value

  param_settings[4].limited[0] = 1;	// lower bound active
  param_settings[4].limits[0] = -0.5;	// lower bound value
  param_settings[4].limited[1] = 1;	// upper bound active
  param_settings[4].limits[1] = 0.5;	// upper bound value
  param_settings[4].advlimited = 1;	// limit advance per iteration
  param_settings[4].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
