// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_anglstretch_xy321shifted;
// DECL: distortion_initfunc_t distortion_anglstretch_xy321shifted_init;
// REG_DISTORTION: {"anglstretch_xy321shifted", 15, distortion_anglstretch_xy321shifted, distortion_anglstretch_xy321shifted_init}

// First 3-param stretch in any angular direction from shifted center
// (cylindrical distortion)
// then 321 radial distortion separately for x and y from (other)
// shifted centers.

void
distortion_anglstretch_xy321shifted (double officialx, double officialy,
				     double *distortedx, double *distortedy,
				     double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      {
	double a = params[0];
	double b = params[1];
	double c = params[2];
	double theta = params[3];
	double centerx = params[4];
	double centery = 0;	// No effect! theta line only shifts <->
	double shiftedx, shiftedy;
	double rotx, roty;

	// First shift
	shiftedx = officialx - centerx;
	shiftedy = officialy - centery;

	// Then rotate x and y
	// http://planning.cs.uiuc.edu/node98.html
	rotx = shiftedx * cos (theta) - shiftedy * sin (theta);
	roty = shiftedx * sin (theta) + shiftedy * cos (theta);

	rotx = a * rotx * rotx * rotx + b * rotx * rotx + c * rotx;

	// Rotate back
	shiftedx = rotx * cos (-theta) - roty * sin (-theta);
	shiftedy = rotx * sin (-theta) + roty * cos (-theta);

	// And shift back
	officialx = shiftedx + centerx;
	officialy = shiftedy + centery;
      }
      {
	double a = params[5];
	double b = params[6];
	double c = params[7];
	// Distortion center is shifted by params.
	double centerx = params[8];
	double centery = params[9];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (a * r * r * r + b * r * r + c * r + (1 - a - b - c));

	*distortedx = officialx * fact;
	*distortedx += centerx;
      }
      {
	double a = params[10];
	double b = params[11];
	double c = params[12];
	// Distortion center is shifted by params.
	double centerx = params[13];
	double centery = params[14];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (a * r * r * r + b * r * r + c * r + (1 - a - b - c));

	*distortedy = officialy * fact;
	*distortedy += centery;
      }
    }
  while (0);
}

void
distortion_anglstretch_xy321shifted_init (double *param,
					  mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = 1.001;
  param[3] = .001;
  param[4] = .001;

  param[5] = .001;
  param[6] = .001;
  param[7] = .001;
  param[8] = .001;
  param[9] = .001;

  param[10] = .001;
  param[11] = .001;
  param[12] = .001;
  param[13] = .001;
  param[14] = .001;

  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = -0.5;	// lower bound value
  param_settings[0].limited[1] = 1;	// upper bound active
  param_settings[0].limits[1] = 0.5;	// upper bound value
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 0.2;	// advance limit value

  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = -0.5;	// lower bound value
  param_settings[1].limited[1] = 1;	// upper bound active
  param_settings[1].limits[1] = 0.5;	// upper bound value
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 0.2;	// advance limit value

  param_settings[2].limited[0] = 1;	// lower bound active
  param_settings[2].limits[0] = 0.1;	// lower bound value
  param_settings[2].limited[1] = 1;	// upper bound active
  param_settings[2].limits[1] = 10.;	// upper bound value
  param_settings[2].advlimited = 1;	// limit advance per iteration
  param_settings[2].advlimit = 0.2;	// advance limit value

  // Angle is periodical
  param_settings[3].periodic = 1;	// periodicity active
  param_settings[3].period = 2. * M_PI;	// wraps into -period/2 .. period/2
  param_settings[3].advlimited = 1;	// limit advance per iteration
  param_settings[3].advlimit = 0.5;	// advance limit value
  param_settings[3].step = 0.001;

  param_settings[4].limited[0] = 1;	// lower bound active
  param_settings[4].limits[0] = -1.5;	// lower bound value
  param_settings[4].limited[1] = 1;	// upper bound active
  param_settings[4].limits[1] = 1.5;	// upper bound value
  param_settings[4].advlimited = 1;	// limit advance per iteration
  param_settings[4].advlimit = 0.2;	// advance limit value

  param_settings[5].limited[0] = 1;	// lower bound active
  param_settings[5].limits[0] = -0.5;	// lower bound value
  param_settings[5].limited[1] = 1;	// upper bound active
  param_settings[5].limits[1] = 0.5;	// upper bound value
  param_settings[5].advlimited = 1;	// limit advance per iteration
  param_settings[5].advlimit = 0.2;	// advance limit value

  param_settings[6].limited[0] = 1;	// lower bound active
  param_settings[6].limits[0] = -0.5;	// lower bound value
  param_settings[6].limited[1] = 1;	// upper bound active
  param_settings[6].limits[1] = 0.5;	// upper bound value
  param_settings[6].advlimited = 1;	// limit advance per iteration
  param_settings[6].advlimit = 0.2;	// advance limit value

  param_settings[7].limited[0] = 1;	// lower bound active
  param_settings[7].limits[0] = -0.5;	// lower bound value
  param_settings[7].limited[1] = 1;	// upper bound active
  param_settings[7].limits[1] = 0.5;	// upper bound value
  param_settings[7].advlimited = 1;	// limit advance per iteration
  param_settings[7].advlimit = 0.2;	// advance limit value

  param_settings[8].limited[0] = 1;	// lower bound active
  param_settings[8].limits[0] = -0.5;	// lower bound value
  param_settings[8].limited[1] = 1;	// upper bound active
  param_settings[8].limits[1] = 0.5;	// upper bound value
  param_settings[8].advlimited = 1;	// limit advance per iteration
  param_settings[8].advlimit = 0.2;	// advance limit value

  param_settings[9].limited[0] = 1;	// lower bound active
  param_settings[9].limits[0] = -0.5;	// lower bound value
  param_settings[9].limited[1] = 1;	// upper bound active
  param_settings[9].limits[1] = 0.5;	// upper bound value
  param_settings[9].advlimited = 1;	// limit advance per iteration
  param_settings[9].advlimit = 0.2;	// advance limit value

  param_settings[10].limited[0] = 1;	// lower bound active
  param_settings[10].limits[0] = -0.5;	// lower bound value
  param_settings[10].limited[1] = 1;	// upper bound active
  param_settings[10].limits[1] = 0.5;	// upper bound value
  param_settings[10].advlimited = 1;	// limit advance per iteration
  param_settings[10].advlimit = 0.2;	// advance limit value

  param_settings[11].limited[0] = 1;	// lower bound active
  param_settings[11].limits[0] = -0.5;	// lower bound value
  param_settings[11].limited[1] = 1;	// upper bound active
  param_settings[11].limits[1] = 0.5;	// upper bound value
  param_settings[11].advlimited = 1;	// limit advance per iteration
  param_settings[11].advlimit = 0.2;	// advance limit value

  param_settings[12].limited[0] = 1;	// lower bound active
  param_settings[12].limits[0] = -0.5;	// lower bound value
  param_settings[12].limited[1] = 1;	// upper bound active
  param_settings[12].limits[1] = 0.5;	// upper bound value
  param_settings[12].advlimited = 1;	// limit advance per iteration
  param_settings[12].advlimit = 0.2;	// advance limit value

  param_settings[13].limited[0] = 1;	// lower bound active
  param_settings[13].limits[0] = -0.5;	// lower bound value
  param_settings[13].limited[1] = 1;	// upper bound active
  param_settings[13].limits[1] = 0.5;	// upper bound value
  param_settings[13].advlimited = 1;	// limit advance per iteration
  param_settings[13].advlimit = 0.2;	// advance limit value

  param_settings[14].limited[0] = 1;	// lower bound active
  param_settings[14].limits[0] = -0.5;	// lower bound value
  param_settings[14].limited[1] = 1;	// upper bound active
  param_settings[14].limits[1] = 0.5;	// upper bound value
  param_settings[14].advlimited = 1;	// limit advance per iteration
  param_settings[14].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
