// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_opencv40;
// DECL: distortion_initfunc_t distortion_opencv40_init;
// REG_DISTORTION: {"opencv40", 12, distortion_opencv40, distortion_opencv40_init}

// model according to openCV 4.0
// https://docs.opencv.org/master/d9/d0c/group__calib3d.html

void
distortion_opencv40 (double officialx, double officialy, double *distortedx,
		     double *distortedy, double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are equal in x and y direction.
      // Distortion center is fixed at 0,0 in this model.
      double k1 = params[0];
      double k2 = params[1];
      double k3 = params[2];
      double k4 = params[3];
      double k5 = params[4];
      double k6 = params[5];
      double p1 = params[6];
      double p2 = params[7];
      double s1 = params[8];
      double s2 = params[9];
      double s3 = params[10];
      double s4 = params[11];

      double r2 = officialx * officialy;
      double fact1;

      fact1 = (1. + k1 * r2 + k2 * r2 * r2 + k3 * r2 * r2 * r2) /
	(1. + k4 * r2 + k5 * r2 * r2 + k6 * r2 * r2 * r2);

      *distortedx = officialx * fact1
	+ 2. * p1 * officialx * officialy
	+ p2 * (r2 + 2. * officialx * officialx) + s1 * r2 + s2 * r2 * r2;

      *distortedy = officialy * fact1
	+ p1 * (r2 + 2. * officialy * officialy)
	+ 2. * p2 * officialx * officialy + s3 * r2 + s4 * r2 * r2;
    }
  while (0);
}

void
distortion_opencv40_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;
  param[3] = .001;
  param[4] = .001;
  param[5] = .001;
  param[6] = .001;
  param[7] = .001;
  param[8] = .001;
  param[9] = .001;
  param[10] = .001;
  param[11] = .001;

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
