// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_opencv40shifted;
// DECL: distortion_initfunc_t distortion_opencv40shifted_init;
// REG_DISTORTION: {"opencv40shifted", 14, distortion_opencv40shifted, distortion_opencv40shifted_init}

void
distortion_opencv40shifted (double officialx, double officialy,
			    double *distortedx, double *distortedy,
			    double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are equal in x and y direction.
      double k1 = params[0];
      double k2 = params[1];
      double k3 = params[2];
      double k4 = params[3];
      double k5 = params[4];
      double k6 = params[5];
      double p1 = params[6];
      double p2 = params[7];
      double s1 = params[8];
      double s2 = params[9];
      double s3 = params[10];
      double s4 = params[11];
      // Distortion center is shifted by params.
      double centerx = params[12];
      double centery = params[13];
      double shiftedx = officialx - centerx;
      double shiftedy = officialy - centery;

      double r2 = shiftedx * shiftedy;
      double fact1;

      fact1 = (1. + k1 * r2 + k2 * r2 * r2 + k3 * r2 * r2 * r2) /
	(1. + k4 * r2 + k5 * r2 * r2 + k6 * r2 * r2 * r2);

      *distortedx = shiftedx * fact1
	+ 2. * p1 * shiftedx * shiftedy
	+ p2 * (r2 + 2. * shiftedx * shiftedx) + s1 * r2 + s2 * r2 * r2;

      *distortedy = shiftedy * fact1
	+ p1 * (r2 + 2. * shiftedy * shiftedy)
	+ 2. * p2 * shiftedx * shiftedy + s3 * r2 + s4 * r2 * r2;

      *distortedx += centerx;
      *distortedy += centery;
    }
  while (0);
}

void
distortion_opencv40shifted_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;
  param[3] = .001;
  param[4] = .001;
  param[5] = .001;
  param[6] = .001;
  param[7] = .001;
  param[8] = .001;
  param[9] = .001;
  param[10] = .001;
  param[11] = .001;
  param[12] = .001;
  param[13] = .001;

  param_settings[12].limited[0] = 1;	// lower bound active
  param_settings[12].limits[0] = -0.5;	// lower bound value
  param_settings[12].limited[1] = 1;	// upper bound active
  param_settings[12].limits[1] = 0.5;	// upper bound value
  param_settings[12].advlimited = 1;	// limit advance per iteration
  param_settings[12].advlimit = 0.2;	// advance limit value

  param_settings[13].limited[0] = 1;	// lower bound active
  param_settings[13].limits[0] = -0.5;	// lower bound value
  param_settings[13].limited[1] = 1;	// upper bound active
  param_settings[13].limits[1] = 0.5;	// upper bound value
  param_settings[13].advlimited = 1;	// limit advance per iteration
  param_settings[13].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
