// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_xy1div12shifted;
// DECL: distortion_initfunc_t distortion_xy1div12shifted_init;
// REG_DISTORTION: {"xy1div12shifted", 10, distortion_xy1div12shifted, distortion_xy1div12shifted_init}

void
distortion_xy1div12shifted (double officialx, double officialy,
			    double *distortedx, double *distortedy,
			    double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are completely separate in x and y direction.
      {
	double a = params[0];
	double b = params[1];
	double c = params[2];
	// Distortion center is shifted by params.
	double centerx = params[3];
	double centery = params[4];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (1 + a * r) / (1 + b * r + c * r * r);

	*distortedx = officialx * fact;
	*distortedx += centerx;
      }
      {
	double a = params[5];
	double b = params[6];
	double c = params[7];
	// Distortion center is shifted by params.
	double centerx = params[8];
	double centery = params[9];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (1 + a * r) / (1 + b * r + c * r * r);

	*distortedy = officialy * fact;
	*distortedy += centery;
      }
    }
  while (0);
}

void
distortion_xy1div12shifted_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;
  param[3] = .001;
  param[4] = .001;

  param[5] = .001;
  param[6] = .001;
  param[7] = .001;
  param[8] = .001;
  param[9] = .001;

  param_settings[3].limited[0] = 1;	// lower bound active
  param_settings[3].limits[0] = -0.5;	// lower bound value
  param_settings[3].limited[1] = 1;	// upper bound active
  param_settings[3].limits[1] = 0.5;	// upper bound value
  param_settings[3].advlimited = 1;	// limit advance per iteration
  param_settings[3].advlimit = 0.2;	// advance limit value

  param_settings[4].limited[0] = 1;	// lower bound active
  param_settings[4].limits[0] = -0.5;	// lower bound value
  param_settings[4].limited[1] = 1;	// upper bound active
  param_settings[4].limits[1] = 0.5;	// upper bound value
  param_settings[4].advlimited = 1;	// limit advance per iteration
  param_settings[4].advlimit = 0.2;	// advance limit value

  param_settings[8].limited[0] = 1;	// lower bound active
  param_settings[8].limits[0] = -0.5;	// lower bound value
  param_settings[8].limited[1] = 1;	// upper bound active
  param_settings[8].limits[1] = 0.5;	// upper bound value
  param_settings[8].advlimited = 1;	// limit advance per iteration
  param_settings[8].advlimit = 0.2;	// advance limit value

  param_settings[9].limited[0] = 1;	// lower bound active
  param_settings[9].limits[0] = -0.5;	// lower bound value
  param_settings[9].limited[1] = 1;	// upper bound active
  param_settings[9].limits[1] = 0.5;	// upper bound value
  param_settings[9].advlimited = 1;	// limit advance per iteration
  param_settings[9].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
