// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "camfuncs.h"
#include <math.h>

// DECL: distortion_func_t distortion_xy4321shifted;
// DECL: distortion_initfunc_t distortion_xy4321shifted_init;
// REG_DISTORTION: {"xy4321shifted", 12, distortion_xy4321shifted, distortion_xy4321shifted_init}

void
distortion_xy4321shifted (double officialx, double officialy,
			  double *distortedx, double *distortedy,
			  double *params)
{
  do
    {
      // Force loop to hopefully get maximal optimizations

      // Params in this model are completely separate in x and y direction.
      {
	double a = params[0];
	double b = params[1];
	double c = params[2];
	double d = params[3];
	// Distortion center is shifted by params.
	double centerx = params[4];
	double centery = params[5];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (a * r * r * r * r + b * r * r * r + c * r * r + d * r +
		(1 - a - b - c - d));

	*distortedx = officialx * fact;
	*distortedx += centerx;
      }
      {
	double a = params[6];
	double b = params[7];
	double c = params[8];
	double d = params[9];
	// Distortion center is shifted by params.
	double centerx = params[10];
	double centery = params[11];
	double shiftedx = officialx - centerx;
	double shiftedy = officialy - centery;

	double r = hypot (shiftedx, shiftedy);
	double fact;

	fact = (a * r * r * r * r + b * r * r * r + c * r * r + d * r +
		(1 - a - b - c - d));

	*distortedy = officialy * fact;
	*distortedy += centery;
      }
    }
  while (0);
}

void
distortion_xy4321shifted_init (double *param, mp_par * param_settings)
{
  param[0] = .001;
  param[1] = .001;
  param[2] = .001;
  param[3] = .001;
  param[4] = .001;
  param[5] = .001;

  param[6] = .001;
  param[7] = .001;
  param[8] = .001;
  param[9] = .001;
  param[10] = .001;
  param[11] = .001;

  param_settings[0].limited[0] = 1;	// lower bound active
  param_settings[0].limits[0] = -0.5;	// lower bound value
  param_settings[0].limited[1] = 1;	// upper bound active
  param_settings[0].limits[1] = 0.5;	// upper bound value
  param_settings[0].advlimited = 1;	// limit advance per iteration
  param_settings[0].advlimit = 0.2;	// advance limit value

  param_settings[1].limited[0] = 1;	// lower bound active
  param_settings[1].limits[0] = -0.5;	// lower bound value
  param_settings[1].limited[1] = 1;	// upper bound active
  param_settings[1].limits[1] = 0.5;	// upper bound value
  param_settings[1].advlimited = 1;	// limit advance per iteration
  param_settings[1].advlimit = 0.2;	// advance limit value

  param_settings[2].limited[0] = 1;	// lower bound active
  param_settings[2].limits[0] = -0.5;	// lower bound value
  param_settings[2].limited[1] = 1;	// upper bound active
  param_settings[2].limits[1] = 0.5;	// upper bound value
  param_settings[2].advlimited = 1;	// limit advance per iteration
  param_settings[2].advlimit = 0.2;	// advance limit value

  param_settings[3].limited[0] = 1;	// lower bound active
  param_settings[3].limits[0] = -0.5;	// lower bound value
  param_settings[3].limited[1] = 1;	// upper bound active
  param_settings[3].limits[1] = 0.5;	// upper bound value
  param_settings[3].advlimited = 1;	// limit advance per iteration
  param_settings[3].advlimit = 0.2;	// advance limit value

  param_settings[4].limited[0] = 1;	// lower bound active
  param_settings[4].limits[0] = -0.5;	// lower bound value
  param_settings[4].limited[1] = 1;	// upper bound active
  param_settings[4].limits[1] = 0.5;	// upper bound value
  param_settings[4].advlimited = 1;	// limit advance per iteration
  param_settings[4].advlimit = 0.2;	// advance limit value

  param_settings[5].limited[0] = 1;	// lower bound active
  param_settings[5].limits[0] = -0.5;	// lower bound value
  param_settings[5].limited[1] = 1;	// upper bound active
  param_settings[5].limits[1] = 0.5;	// upper bound value
  param_settings[5].advlimited = 1;	// limit advance per iteration
  param_settings[5].advlimit = 0.2;	// advance limit value

  param_settings[6].limited[0] = 1;	// lower bound active
  param_settings[6].limits[0] = -0.5;	// lower bound value
  param_settings[6].limited[1] = 1;	// upper bound active
  param_settings[6].limits[1] = 0.5;	// upper bound value
  param_settings[6].advlimited = 1;	// limit advance per iteration
  param_settings[6].advlimit = 0.2;	// advance limit value

  param_settings[7].limited[0] = 1;	// lower bound active
  param_settings[7].limits[0] = -0.5;	// lower bound value
  param_settings[7].limited[1] = 1;	// upper bound active
  param_settings[7].limits[1] = 0.5;	// upper bound value
  param_settings[7].advlimited = 1;	// limit advance per iteration
  param_settings[7].advlimit = 0.2;	// advance limit value

  param_settings[8].limited[0] = 1;	// lower bound active
  param_settings[8].limits[0] = -0.5;	// lower bound value
  param_settings[8].limited[1] = 1;	// upper bound active
  param_settings[8].limits[1] = 0.5;	// upper bound value
  param_settings[8].advlimited = 1;	// limit advance per iteration
  param_settings[8].advlimit = 0.2;	// advance limit value

  param_settings[9].limited[0] = 1;	// lower bound active
  param_settings[9].limits[0] = -0.5;	// lower bound value
  param_settings[9].limited[1] = 1;	// upper bound active
  param_settings[9].limits[1] = 0.5;	// upper bound value
  param_settings[9].advlimited = 1;	// limit advance per iteration
  param_settings[9].advlimit = 0.2;	// advance limit value

  param_settings[10].limited[0] = 1;	// lower bound active
  param_settings[10].limits[0] = -0.5;	// lower bound value
  param_settings[10].limited[1] = 1;	// upper bound active
  param_settings[10].limits[1] = 0.5;	// upper bound value
  param_settings[10].advlimited = 1;	// limit advance per iteration
  param_settings[10].advlimit = 0.2;	// advance limit value

  param_settings[11].limited[0] = 1;	// lower bound active
  param_settings[11].limits[0] = -0.5;	// lower bound value
  param_settings[11].limited[1] = 1;	// upper bound active
  param_settings[11].limits[1] = 0.5;	// upper bound value
  param_settings[11].advlimited = 1;	// limit advance per iteration
  param_settings[11].advlimit = 0.2;	// advance limit value

  // Note: don't forget adjusting #params in REG_DISTORTION above
}
