#ifndef COMMON_H
#define COMMON_H

#include <unistd.h>
#include <mosquitto.h>
#include <jansson.h>
#include "../shared/mqttpub.h"

//#define INTERP_DATA_KEEP_ALLOCD

#define DEBUG

#ifdef DEBUG
#define MDBG(format, args...)  \
    mqttprintf(state->mosq,  \
        state->topicprefix, "collect", state->name, "debug", NULL,  \
        "%s %d %s:%d %s(): " format ,  \
        progname, getpid(), __FILE__, __LINE__, __func__, ##args)
#else
#define MDBG(format, args...)  do { } while(0)
#endif


typedef struct
{
  int fromx;
  int fromy;
  double weight;
} interp_data_t;

typedef struct
{
  struct mosquitto *mosq;

  char *svr_hostname;
  int svr_port;
  int connstate;
  time_t connect_nexttime;	/* Planned time of next connect attempt;
				   also stdin-EOF force-disconnect time */
  char *configfile;
  json_t *jconfig;
  json_t *jaxes[2];		// fixed array
  json_t **jplanes;		// malloc'd array
  int jplanes_num;
  json_t *jconditions;
  json_t *jrenders;

  char *topicprefix;
  const char *name;
  const char *data_topic;

  char broadcmd_topic[TOPICLEN + 5];
  char collectcmd_topic[TOPICLEN + 5];
  char ourcmd_topic[TOPICLEN + 5];

  int do_interpolate;
  int interpolate_debug;

  double xbinsize;
  double ybinsize;

  double lasttime_d;		// for autogenerated:time plane
  double lastrecv_x;		// last-received-position marker
  double lastrecv_y;

  double *data;			// mallocd, sometimes *replaced*
  int data_valuesperbin;	// ==jplanes_num, or *2 when interpolate
  int data_xsize;		// how much is mallocd
  int data_ysize;
  int data_xzerooffset;		// 0,0 is at this x,y offset in data[]
  int data_yzerooffset;
  int data_xused_min;		// bin coords (not data[] index)
  int data_xused_max;
  int data_yused_min;
  int data_yused_max;

  interp_data_t *interp_data;
  int interp_data_valuesperbin;

} state_t;

extern const char *progname;

#endif /* COMMON_H */
