// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include <string.h>

#include "data_test.h"
#include "data.h"
#include "../shared/error.h"

#define EXPECT(c)					\
  do {							\
    if (!(c)) {						\
      STDERR("Test fail: %s is FALSE", #c);		\
      testresult = 1;			 		\
    }							\
  } while(0) /*END*/
#define EXPECT_EQ(a, b)					\
  do {							\
    if ((a) != (b)) {					\
      STDERR("Test fail: %s = %f is not equal to %s = %f", \
	#a, (double) (a), #b, (double) (b));		\
      testresult = 1;			 		\
    }							\
  } while(0) /*END*/
#define DUMMY_to_keep_indent_happy(x)	/* nothing */
DUMMY_to_keep_indent_happy (x);

int
data_test ()
{
  int testresult = 0;
  state_t state;

  memset (&state, 0, sizeof (state));
  state.data_xused_max = -1;	// prevent getdata() errors
  state.data_yused_max = -1;
  state.data_valuesperbin = 2;

  EXPECT (isnan (getdata (&state, 0, 0, 0)));
  EXPECT (isnan (getdata (&state, 100, 100, 0)));

  // Store initial point at (100,200) to give "shifted" allocation
  storedata (&state, 100, 200, 0, 1.);
  storedata (&state, 100, 200, 1, 2.);

  EXPECT_EQ (getdata (&state, 100, 200, 0), 1.);
  EXPECT_EQ (getdata (&state, 100, 200, 1), 2.);

  EXPECT_EQ (state.data_xsize, 20);
  EXPECT_EQ (state.data_ysize, 20);
  EXPECT_EQ (state.data_xzerooffset, -90);
  EXPECT_EQ (state.data_yzerooffset, -190);
  EXPECT_EQ (state.data_xused_min, 100);
  EXPECT_EQ (state.data_xused_max, 100);
  EXPECT_EQ (state.data_yused_min, 200);
  EXPECT_EQ (state.data_yused_max, 200);
  EXPECT_EQ (state.data[(10 * 20 + 10) * 2 + 1], 2.);

  // Store more towards (104,204) to give some "body", no realloc
  storedata (&state, 104, 200, 0, 3.);
  storedata (&state, 100, 204, 1, 4.);

  EXPECT_EQ (getdata (&state, 104, 200, 0), 3.);
  EXPECT_EQ (getdata (&state, 100, 204, 1), 4.);

  EXPECT_EQ (state.data_xsize, 20);
  EXPECT_EQ (state.data_ysize, 20);
  EXPECT_EQ (state.data_xzerooffset, -90);
  EXPECT_EQ (state.data_yzerooffset, -190);
  EXPECT_EQ (state.data_xused_min, 100);
  EXPECT_EQ (state.data_xused_max, 104);
  EXPECT_EQ (state.data_yused_min, 200);
  EXPECT_EQ (state.data_yused_max, 204);

  // Store far away at (400,200/202) to force realloc with loss of
  // unused entries
  storedata (&state, 400, 200, 0, 5.);
  storedata (&state, 400, 202, 1, 6.);

  EXPECT_EQ (getdata (&state, 400, 200, 0), 5.);
  EXPECT_EQ (getdata (&state, 400, 202, 1), 6.);

  // Check old data still present
  EXPECT_EQ (getdata (&state, 100, 200, 0), 1.);
  EXPECT_EQ (getdata (&state, 100, 200, 1), 2.);
  EXPECT_EQ (getdata (&state, 104, 200, 0), 3.);
  EXPECT_EQ (getdata (&state, 100, 204, 1), 4.);

  EXPECT_EQ (state.data_xsize, 602);	// 100..400 = 301 entries, *2 growth
  EXPECT_EQ (state.data_ysize, 10);
  EXPECT_EQ (state.data_xzerooffset, 50);	// -50..551 allocd
  EXPECT_EQ (state.data_yzerooffset, -198);	// 198..207 allocd
  EXPECT_EQ (state.data_xused_min, 100);
  EXPECT_EQ (state.data_xused_max, 400);
  EXPECT_EQ (state.data_yused_min, 200);
  EXPECT_EQ (state.data_yused_max, 204);

  if (testresult == 0)
    printf ("%s: all tests successful\n", __FUNCTION__);

  return testresult;
}
