#ifndef COMMON_H
#define COMMON_H

#include <unistd.h>
#include <mosquitto.h>
#include <mgl2/mgl_cf.h>
#include "../shared/mqttpub.h"

#define FNAMELEN  10000

#define DEBUG

#ifdef DEBUG
#define MDBG(format, args...)  \
    mqttprintf(state->mosq,  \
        state->topicprefix, "render", state->name, "debug", NULL,  \
        "%s %d %s:%d %s(): " format ,  \
        progname, getpid(), __FILE__, __LINE__, __func__, ##args)
#else
#define MDBG(format, args...)  do { } while(0)
#endif


typedef struct
{
  struct mosquitto *mosq;

  char *svr_hostname;
  int svr_port;
  int connstate;
  time_t connect_nexttime;	/* Planned time of next connect attempt;
				   also stdin-EOF force-disconnect time */

  char *topicprefix;
  char *name;
  char name_sane[TOPICLEN + 5];
  char *pubdir;
  char *pubpfx;
  char **recv_topics;
  int recv_topics_num;

  char broadcmd_topic[TOPICLEN + 5];
  char rendercmd_topic[TOPICLEN + 5];
  char ourcmd_topic[TOPICLEN + 5];

  int overactive_counter;
  int idletimer_fd;
  int idletimer_running;
} state_t;

typedef struct
{
  const char *label;
  int items;
  double interval;
  double firstitem;

  int mgl_items;
  int mgl_extra_items_at_front;
  HMDT mgl_list;
  HMDT mgl_list_minushalf_1extra;
} axis_t;

typedef struct
{
  HMDT mdata;
  mreal mmin;
  mreal mmax;
  double filled_factor;
} mathgl_plane_t;

typedef struct
{
  const char *label;
  double rounddata;		// Per plane: applied when filling mgl_plane
  int is_autogen_time;
  double time_last;
  double time_now;

  mathgl_plane_t mathgl_plane;
  mathgl_plane_t mathgl_plane_1extra;
} plane_t;

typedef struct
{
  int valid;
  double x;
  double y;
} lastreceived_t;

typedef struct
{
  char collecttopic[TOPICLEN + 10];
  char rendername_sane[TOPICLEN + 10];

  double *data;
  axis_t axes[2];
  plane_t *planes;		// mallocd
  int planes_num;
  lastreceived_t lastreceived;
} data_t;

extern const char *progname;

#endif /* COMMON_H */
