// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "mathgl_prepare.h"
#include "../shared/error.h"

void
mathgl_init_axes (state_t * state, data_t * data)
{
  int ax;

  for (ax = 0; ax < 2; ax++)
    {
      double lastitem;
      int extra_items_at_end;
      int j;
      mreal *mgldataptr;

      if (data->axes[ax].mgl_list != NULL)
	continue;		// Already done

      // Extend axes to have origin displayed.
      // (Will end up using NAN data = transparency.)

      //               2.2 3.2    axis: items=2 firstitem=2.2 interval=1.0
      //               [0] [1]
      //  -0.8 0.2 1.2 2.2 3.2    mgl: extra=3 items=5 first=-0.8
      //  [0]  [1] [2] [3] [4]

      data->axes[ax].mgl_items = data->axes[ax].items;

      data->axes[ax].mgl_extra_items_at_front =
	lrint (ceil (data->axes[ax].firstitem / data->axes[ax].interval));
      if (data->axes[ax].mgl_extra_items_at_front < 0)
	data->axes[ax].mgl_extra_items_at_front = 0;

      data->axes[ax].mgl_items += data->axes[ax].mgl_extra_items_at_front;

      //  -3.6 -2.6                 axis: items=2 firstitem=-3.6 interval=1.0
      //   [0]  [1]
      //  -3.6 -2.6 -1.6 -0.6 0.4   mgl: extra=3 items=5
      //   [0]  [1]  [2]  [3] [4]

      lastitem = data->axes[ax].firstitem +
	data->axes[ax].items * data->axes[ax].interval;

      extra_items_at_end = lrint (ceil (-lastitem / data->axes[ax].interval));
      if (extra_items_at_end < 0)
	extra_items_at_end = 0;
      data->axes[ax].mgl_items += extra_items_at_end;

#if 0
      STDERR ("mgl axis[%d] items %d mgl_items %d front %d",
	      ax, data->axes[ax].items,
	      data->axes[ax].mgl_items,
	      data->axes[ax].mgl_extra_items_at_front);
#endif

      // Fill lists

      data->axes[ax].mgl_list =
	mgl_create_data_size (data->axes[ax].mgl_items, 1, 1);
      mgldataptr = mgl_data_data (data->axes[ax].mgl_list);
      for (j = 0; j < data->axes[ax].mgl_items; j++)
	mgldataptr[j] =
	  ((double) (j - data->axes[ax].mgl_extra_items_at_front)) *
	  data->axes[ax].interval + data->axes[ax].firstitem;

      data->axes[ax].mgl_list_minushalf_1extra =
	mgl_create_data_size (data->axes[ax].mgl_items + 1, 1, 1);
      mgldataptr = mgl_data_data (data->axes[ax].mgl_list_minushalf_1extra);
      for (j = 0; j < data->axes[ax].mgl_items + 1; j++)
	mgldataptr[j] =
	  (((double) (j - data->axes[ax].mgl_extra_items_at_front)) -
	   0.5) * data->axes[ax].interval + data->axes[ax].firstitem;
    }
}

static inline double
getdata_mgl (data_t * data, int mgl_x, int mgl_y, int planeidx,
	     long *items_nonNAN)
{
  int data_x = mgl_x - data->axes[0].mgl_extra_items_at_front;
  int data_y = mgl_y - data->axes[1].mgl_extra_items_at_front;
  double d;

  if (data_x < 0 || data_x >= data->axes[0].items)
    return NAN;
  if (data_y < 0 || data_y >= data->axes[1].items)
    return NAN;

  d = data->data[(data_y * data->axes[0].items + data_x)
		 * data->planes_num + planeidx];

  if (!isnan (d))
    (*items_nonNAN)++;

  return d;
}

void
fillmatrix (state_t * state, data_t * data, int planeidx,
	    mathgl_plane_t * mgl_plane, int itemsx, int itemsy)
{
  int x, y;
  double d;
  long items_nonNAN = 0;
  mreal *mgldataptr = mgl_data_data (mgl_plane->mdata);
  double rounddata = data->planes[planeidx].rounddata;

  for (y = 0; y < itemsy; y++)
    for (x = 0; x < itemsx; x++)
      {
	d = getdata_mgl (data, x, y, planeidx, &items_nonNAN);

	if (data->planes[planeidx].is_autogen_time && !isnan (d))
	  {
	    // Rewrite time in minutes-ago, up to 1 hour.
	    d = (d - data->planes[planeidx].time_now) / 60.;

	    // Colorbar top is at ceil(max+0.02),
	    // make sure we're always well below zero.
	    d -= 0.05;

	    if (d < -60.)
	      d = NAN;
	  }

	if (rounddata > 0.000001)
	  d = round (d / rounddata) * rounddata;

	mgldataptr[y * itemsx + x] = d;
      }

  mgl_plane->mmin = mgl_data_min (mgl_plane->mdata);
  mgl_plane->mmax = mgl_data_max (mgl_plane->mdata);

  mgl_plane->filled_factor = ((double) items_nonNAN) /
    ((double) (data->axes[0].items * data->axes[1].items));

#if 0
  STDERR
    ("filled %d x %d, axes %d x %d, mgl_axes %d x %d front %d %d, nonNAN %ld",
     itemsx, itemsy, data->axes[0].items, data->axes[1].items,
     data->axes[0].mgl_items, data->axes[1].mgl_items,
     data->axes[0].mgl_extra_items_at_front,
     data->axes[1].mgl_extra_items_at_front, items_nonNAN);
#endif
}

void
mathgl_init_plane_normal (state_t * state, data_t * data, int planeidx)
{
  plane_t *plane = &data->planes[planeidx];
  mathgl_plane_t *mgl_plane = &plane->mathgl_plane;

  if (mgl_plane->mdata != NULL)
    return;			// Already done

  mathgl_init_axes (state, data);

  mgl_plane->mdata = mgl_create_data_size (data->axes[0].mgl_items,
					   data->axes[1].mgl_items, 1);

  fillmatrix (state, data, planeidx, mgl_plane,
	      data->axes[0].mgl_items, data->axes[1].mgl_items);
}

void
mathgl_init_plane_1extra (state_t * state, data_t * data, int planeidx)
{
  plane_t *plane = &data->planes[planeidx];
  mathgl_plane_t *mgl_plane = &plane->mathgl_plane_1extra;

  if (mgl_plane->mdata != NULL)
    return;			// Already done

  mathgl_init_axes (state, data);

  mgl_plane->mdata = mgl_create_data_size (data->axes[0].mgl_items + 1,
					   data->axes[1].mgl_items + 1, 1);

  fillmatrix (state, data, planeidx, mgl_plane,
	      data->axes[0].mgl_items + 1, data->axes[1].mgl_items + 1);
}

void
mathgl_delete_data (state_t * state, data_t * data)
{
  int i;

  for (i = 0; i < 2; i++)
    {
      if (data->axes[i].mgl_list != NULL)
	{
	  mgl_delete_data (data->axes[i].mgl_list);
	  data->axes[i].mgl_list = NULL;
	}
      if (data->axes[i].mgl_list_minushalf_1extra != NULL)
	{
	  mgl_delete_data (data->axes[i].mgl_list_minushalf_1extra);
	  data->axes[i].mgl_list_minushalf_1extra = NULL;
	}
    }

  for (i = 0; i < data->planes_num; i++)
    {
      if (data->planes[i].mathgl_plane.mdata != NULL)
	{
	  mgl_delete_data (data->planes[i].mathgl_plane.mdata);
	  data->planes[i].mathgl_plane.mdata = NULL;
	}
      if (data->planes[i].mathgl_plane_1extra.mdata != NULL)
	{
	  mgl_delete_data (data->planes[i].mathgl_plane_1extra.mdata);
	  data->planes[i].mathgl_plane_1extra.mdata = NULL;
	}
    }
}
