// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

#include "render_csv.h"
#include "recv_pub.h"

void
render_csv (state_t * state, data_t * data, json_t * jrender,
	    const char *rendername, const char *rendername_sane)
{
  const char *planelabel;
  int planeidx;
  int x, y;
  char destfname[FNAMELEN + 10];
  char tmpfname[FNAMELEN + 10];
  FILE *f;

  planelabel = json_string_value (json_object_get (jrender, "plane"));
  if (planelabel == NULL)
    {
      MDBG ("render object has no plane label");
      return;
    }

  planeidx = lookup_plane (data, planelabel);
  if (planeidx < 0)
    {
      MDBG ("plane %s not found", planelabel);
      return;
    }

  if (!makefilenames (state, rendername_sane, "csv", destfname, tmpfname))
    return;

  f = fopen (tmpfname, "w");
  if (f == NULL)
    return;

  // Header row with X ticks
  for (x = 0; x < data->axes[0].items; x++)
    fprintf (f, ",%f", data->axes[0].firstitem
	     + ((double) x) * data->axes[0].interval);
  fprintf (f, "\n");

  // Now the full matrix.
  for (y = data->axes[1].items - 1; y >= 0; y--)	// note: top-down
    {
      fprintf (f, "%f", data->axes[1].firstitem
	       + ((double) y) * data->axes[1].interval);
      for (x = 0; x < data->axes[0].items; x++)
	{
	  double d = getdata (data, x, y, planeidx);

	  if (data->planes[planeidx].is_autogen_time && !isnan (d))
	    {
	      // Rewrite time in minutes-ago, up to 1 hour.
	      d = (d - data->planes[planeidx].time_now) / 60.;
	      if (d < -60.)
		d = NAN;
	    }

	  if (isnan (d))
	    fprintf (f, ",");
	  else
	    fprintf (f, ",%f", d);
	}
      fprintf (f, "\n");
    }

  fclose (f);
  if (rename (tmpfname, destfname) != 0)
    return;

  // Size as #columns/#rows, including headers

  announcefile (state, data->collecttopic, rendername, rendername_sane,
		"csv", "csv",
		data->axes[0].items + 1, data->axes[1].items + 1);
}
