// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

// This is shared function where cam_t can be extended.

// Return 0 on okay; anything else via external CAMERR_RETURN
int
parse_cam_json (json_t * jroot, cam_t * cam, int require_typedist)
{
  do
    {
      json_t *jcamtype;
      json_t *jparam;
      const char *name;
      int found;
      int i;
      int p;

      jcamtype = json_object_get (jroot, "type");
      jparam = json_object_get (jcamtype, "param");

      if (jcamtype == NULL)
	{
	  if (require_typedist)
	    CAMERR_RETURN ("No camera type specified");
	  else
	    break;
	}

      name = json_string_value (json_object_get (jcamtype, "name"));
      if (name == NULL)
	{
	  if (require_typedist)
	    CAMERR_RETURN ("No camera type name specified");
	  else
	    break;
	}

      found = 0;
      for (i = 0; camtype[i].name != NULL; i++)
	if (!strcmp (name, camtype[i].name))
	  {
	    cam->camtype = i;

	    cam->typeparam = calloc (camtype[i].nparam, sizeof (double));
	    if (cam->typeparam == NULL)
	      {
		STDERR ("malloc failed");
		exit (1);
	      }

	    found = 1;
	    break;
	  }
      if (!found)
	CAMERR_RETURN ("Camera type %s unknown", name);

      for (p = 0;
	   p < json_array_size (jparam) &&
	   p < camtype[cam->camtype].nparam; p++)
	{
	  cam->typeparam[p] = json_number_value (json_array_get (jparam, p));
	}
      // This may read too few params, calloc() will have set rest to 0.
    }
  while (0);

  do
    {
      json_t *jdist;
      json_t *jparam;
      const char *name;
      int found;
      int i;
      int p;

      jdist = json_object_get (jroot, "distortion");
      jparam = json_object_get (jdist, "param");

      if (jdist == NULL)
	{
	  if (require_typedist)
	    CAMERR_RETURN ("No camera distortion specified");
	  else
	    break;
	}

      name = json_string_value (json_object_get (jdist, "name"));
      if (name == NULL)
	{
	  if (require_typedist)
	    CAMERR_RETURN ("No distortion type name specified");
	  else
	    break;
	}

      found = 0;
      for (i = 0; camdistortion[i].name != NULL; i++)
	if (!strcmp (name, camdistortion[i].name))
	  {
	    cam->camdistortion = i;

	    cam->distortionparam =
	      calloc (camdistortion[i].nparam, sizeof (double));
	    if (cam->distortionparam == NULL)
	      {
		STDERR ("malloc failed");
		exit (1);
	      }

	    found = 1;
	    break;
	  }
      if (!found)
	CAMERR_RETURN ("Cameradistortion %s unknown", name);

      for (p = 0;
	   p < json_array_size (jparam) &&
	   p < camdistortion[cam->camdistortion].nparam; p++)
	{
	  cam->distortionparam[p] =
	    json_number_value (json_array_get (jparam, p));
	}
      // This may read too few params, calloc() will have set rest to 0.
    }
  while (0);

  {
    int minpix;

    cam->w = json_integer_value (json_object_get (jroot, "pixw"));
    cam->h = json_integer_value (json_object_get (jroot, "pixh"));
    if (cam->w < 5 || cam->h < 5)
      CAMERR_RETURN ("Camera width/height (pixels) not specified or useless");

    minpix = cam->w;
    if (cam->h < minpix)
      minpix = cam->h;
    cam->xoffset = ((double) cam->w) / 2.;
    cam->yoffset = ((double) cam->h) / 2.;
    cam->xscale = 2. / (double) minpix;
    cam->yscale = cam->xscale;
  }

  return 0;
}
