// Part of SCWIPS  Copyright (C) 2021 J.A.Bezemer@opensourcepartners.nl  GPL>=3

// MQTT docu:
// https://www.eclipse.org/paho/files/jsdoc/Paho.MQTT.Client.html
// https://www.eclipse.org/paho/files/jsdoc/Paho.MQTT.Message.html

var topicprefix;
var lurkermode = false;
var mqttclient;
var position_text_new = "";
var position_text = "";
var positionresult_text_new = "";
var positionresult_text = "";
var sceneresult_text_new = "";
var sceneresult_text = "";
var scenedetails_text_new = "";
var scenedetails_text = "";
var scenecalc_running = false;

var img_fullscreen = "";	// Has "cam/" or "collect/" prefix.
var requestImgsLoopTimer;
var requestImgsLoopForce = false;
// "Iterator" for last-requested name,
// or forced name if requestImgsLoopForce true.
// Note: has "cam/" or "collect/" prefix.
var requestImgsLoopName = "(none)";
var expectingProcIdle = "";

var cams = { };

var collects = { };

var objs =[];

var dists =[];

var img_campix = new Image ();
var img_campix_url = "img/campix.gif";
img_campix.src = img_campix_url;

var img_calcpix = new Image ();
var img_calcpix_url = "img/backcalcpix.gif";
img_calcpix.src = img_calcpix_url;

var img_obj1 = new Image ();
var img_obj1_url = "img/obj1.gif";
img_obj1.src = img_obj1_url;
var obj1_num = -1;

var img_obj2 = new Image ();
var img_obj2_url = "img/obj2.gif";
img_obj2.src = img_obj2_url;
var obj2_num = -1;


// MQTT connection
//---------------------------------------------------------------------------

function
connectButton ()
{
  var svrname;
  var svrport;
  var connectOpts;

  document.getElementById ("connectformdiv").style.display = "none";
  document.getElementById ("statusline").innerHTML = "Connecting...";

  svrname = document.getElementById ("servername").value;
  svrport = document.getElementById ("serverport").value;
  topicprefix = document.getElementById ("topicprefix").value;
  lurkermode = document.getElementById ("lurkermode").checked;

  mqttclient = new Paho.MQTT.Client (svrname, Number (svrport),
				     "websock" + Math.random ());

  // set callback handlers
  mqttclient.onConnectionLost = onConnectionLost;
  mqttclient.onMessageArrived = onMessageArrived;

  connectOpts = {
	// *INDENT-OFF*
	onSuccess: onConnect,
	timeout: 20,
	keepAliveInterval: 10,
	//reconnect: true,  -- just do manually
	onFailure: onConnectFailed,
	// *INDENT-ON*
  };

  mqttclient.connect (connectOpts);
}

function
onConnectFailed (responseObject)
{
  var msg;

  console.log ("onConnectFailed");

  msg = "Connect Failed";

  if (responseObject.errorCode !== 0)
    msg = msg + " (" + responseObject.errorMessage + ")";

  document.getElementById ("statusline").innerHTML = msg;
  document.getElementById ("connectformdiv").style.display = "block";
}

function
onConnectionLost (responseObject)
{
  var msg;

  console.log ("onConnectionLost");

  msg = "Connection Lost";

  if (responseObject.errorCode !== 0)
    msg = msg + " (" + responseObject.errorMessage + ")";

  document.getElementById ("statusline").innerHTML = msg;
  document.getElementById ("connectformdiv").style.display = "block";
}

function
onConnect ()
{
  var message;

  console.log ("onConnect");

  document.getElementById ("statusline").innerHTML = "";
  document.getElementById ("leftcontent").style.display = "block";
  document.getElementById ("rightcolumn").style.display = "block";
  document.getElementById ("rowchooserdiv").style.display = "block";

  //mqttclient.subscribe ("#");
  //mqttclient.subscribe (topicprefix + "/cmd");  --we do not take commands.
  mqttclient.subscribe (topicprefix + "/objtrack/result");
  mqttclient.subscribe (topicprefix + "/scene/result");
  mqttclient.subscribe (topicprefix + "/scene/details");
  mqttclient.subscribe (topicprefix + "/cam/+/info");
  mqttclient.subscribe (topicprefix + "/cam/+/pix");
  mqttclient.subscribe (topicprefix + "/cam/+/calcpix");
  mqttclient.subscribe (topicprefix + "/cam/+/status");
  mqttclient.subscribe (topicprefix + "/collect/+/info");

  mqttclient.subscribe (topicprefix + "/cam/+/img/+/jpeg/+/loc");
  mqttclient.subscribe (topicprefix + "/cam/+/img/+/png/+/loc");
  mqttclient.subscribe (topicprefix + "/collect/+/img/+/jpeg/+/loc");
  mqttclient.subscribe (topicprefix + "/collect/+/img/+/png/+/loc");

  mqttclient.subscribe (topicprefix + "/imgproc/+/status");
  mqttclient.subscribe (topicprefix + "/render/+/status");

  message = new Paho.MQTT.Message ("Connected");
  message.destinationName = topicprefix + "/scwipsweb/debug";
  mqttclient.send (message);

  mqttmsg = new Paho.MQTT.Message ('{"cmd": "postinfo"}');
  mqttmsg.destinationName = topicprefix + "/cam/cmd";
  mqttclient.send (mqttmsg);

  mqttmsg = new Paho.MQTT.Message ('{"cmd": "postresult"}');
  mqttmsg.destinationName = topicprefix + "/scene/cmd";
  mqttclient.send (mqttmsg);

  mqttmsg = new Paho.MQTT.Message ('{"cmd": "postinfo"}');
  mqttmsg.destinationName = topicprefix + "/collect/cmd";
  mqttclient.send (mqttmsg);
}

function
onMessageArrived (message)
{
  var subtopics;
  var topicsplit;

  //console.log ("onMessageArrived to " + message.destinationName + " : " +
  //           message.payloadString);

  if (!message.destinationName.startsWith (topicprefix + "/"))
    return;

  // scwips/cam/c0/info --> cam/c0/info
  subtopics = message.destinationName.replace (topicprefix + "/", "");

  // cam/c0/info --> ["cam", "c0", "info"]
  topicsplit = subtopics.split ("/");
  //console.log ("topicsplit: " + topicsplit);

  if (topicsplit[0] == "cam")
    {
      handleCamMessage (topicsplit[1], topicsplit[2], topicsplit.slice (3),
			message.payloadString);
    }
  else if (topicsplit[0] == "imgproc")
    {
      handleImgprocMessage (topicsplit[1], topicsplit[2],
			    message.payloadString);
    }
  else if (topicsplit[0] == "collect")
    {
      handleCollectMessage (topicsplit[1], topicsplit[2],
			    topicsplit.slice (3), message.payloadString);
    }
  else if (topicsplit[0] == "render")
    {
      handleRenderMessage (topicsplit[1], topicsplit[2],
			   message.payloadString);
    }
  else if (subtopics == "objtrack/result")
    {
      receiveObjtrack (message.payloadString);
    }
  else if (subtopics == "scene/result")
    {
      if (message.payloadString.length >= 1)
	{
	  sceneresult_text_new = "&nbsp;<br/>Scene calculation result:";

	  sceneresult_text_new += "</br><pre>"
	    + escapeHtml (message.payloadString) + "</pre>";

	  scenecalc_running = false;
	}
    }
  else if (subtopics == "scene/result/detailsUNUSED")
    {
      if (message.payloadString.length >= 1)
	scenedetails_text_new +=
	  "<br/><pre>" + escapeHtml (message.payloadString) + "</pre>";
    }
}

function
postanythingButton ()
{
  var message;

  message =
    new Paho.MQTT.Message (document.getElementById ("postanythingmessage").
			   value);

  message.destinationName =
    document.getElementById ("postanythingtopic").value;

  if (message.destinationName.length > 0)
    mqttclient.send (message);
}

function
receiveObjtrack (msgstring)
{
  var jmsg;
  var lostcause;

  try
  {
    jmsg = JSON.parse (msgstring);
  }
  catch (error)
  {
    return;
  }

  window.clearTimeout (objcoordInvalidTimer);

  if (jmsg.hasOwnProperty ("x") &&
      jmsg.hasOwnProperty ("y") &&
      jmsg.hasOwnProperty ("z") &&
      Number.isFinite (jmsg["x"]) &&
      Number.isFinite (jmsg["y"]) && Number.isFinite (jmsg["z"]))
    {
      setPositionText ("<span style=\"background: #00ee00\">"
		       + "Position " + (+jmsg["x"]).toFixed (2)
		       + " " + (+jmsg["y"]).toFixed (2)
		       + " " + (+jmsg["z"]).toFixed (2) + "</span>");
      // (+..) will force to numeric

      // objtrack will report explicit fix-lost in <2sec
      objcoordInvalidTimer = window.setTimeout (invalidateObjcoord, 4000);

      positionresult_text_new = "&nbsp;<br/>Position result:<br/><pre>"
	+ escapeHtml (msgstring) + "</pre>";
    }
  else
    {
      // Assume lost fix

      lostcause = "";
      if (jmsg.hasOwnProperty ("lost") && jmsg.hasOwnProperty ("cause"))
	lostcause = " (" + escapeHtml (jmsg["cause"]) + ")";

      setPositionText ("<span style=\"background: #ff0000\">"
		       + "Position fix lost" + lostcause + "</span>");
      // Already invalid, don't restart timer.
    }
}

var objcoordInvalidTimer = window.setTimeout (invalidateObjcoord, 100000000);
// 100000sec = 27.7h -- shouldn't really trigger before scenecalc is done.

// Timer-triggered invalidation
function
invalidateObjcoord ()
{
  window.clearTimeout (objcoordInvalidTimer);
  setPositionText ("<span style=\"background: #ff0000\">"
		   + "Position calc timeout</span>");
}

function
shutdownCmd ()
{
  var message;

  if (!window.confirm ("Really shutdown everyone?"))
    return;

  message = new Paho.MQTT.Message ('{"cmd": "shutdown"}');
  message.destinationName = topicprefix + "/cmd";
  mqttclient.send (message);
}

// Called from click handler or re-called when Rejected (via timeout)
function
submitScenecalcDo ()
{
  var msg;
  var msgadd;
  var camNames;
  var cami;
  var camname;
  var fval;
  var obji;
  var disti;
  var obj1str;
  var obj2str;
  var mqttmsg;

  msg =[];
  camNames = Object.getOwnPropertyNames (cams);

  fval = parseFloat (document.getElementById ("sumsquareslimit").value);
  if (!isNaN (fval))
    {
      msgadd = {
      "sumsquares-limit":fval
      };
      msg.push (msgadd);
    }

  // Add refobject camera pixels and heights
  for (obji = 0; obji < objs.length; obji++)
    {
      if (objs[obji]["enabled"] != 1)
	continue;

      // Note: scenecalc handles first-mentioned object as origin (0,0,z)
      // and second-mentioned object defines x-axis (x,0,z).
      // So make sure we always introduce objects in correct sequence
      // by starting with camera pixels (should always be >=2 cams)

      for (cami = 0; cami < camNames.length; cami++)
	{
	  camname = camNames[cami];
	  if (Object.prototype.hasOwnProperty.call
	      (objs[obji]["cams"], camname))
	    {
	      msgadd = {
	      "obj":"Refpt_R" + obji,
	      "cam":camname,
	      "pixx":objs[obji]["cams"][camname]["pixx"],
	      "pixy":objs[obji]["cams"][camname]["pixy"],
	      "pixw":objs[obji]["cams"][camname]["pixw"],
	      "pixh":objs[obji]["cams"][camname]["pixh"]
	      };
	      msg.push (msgadd);
	    }
	}

      fval = parseFloat (document.getElementById ("objheight_" + obji).value);
      if (!isNaN (fval))
	{
	  msgadd = {
	  "obj":"Refpt_R" + obji,
	  "z":fval
	  };
	  msg.push (msgadd);
	}
    }

  // Add camera heights
  for (cami = 0; cami < camNames.length; cami++)
    {
      camname = camNames[cami];

      fval =
	parseFloat (document.getElementById ("camheight_" + camname).value);
      if (!isNaN (fval))
	{
	  msgadd = {
	  "obj":camname,
	  "z":fval
	  };
	  msg.push (msgadd);
	}
    }

  // Add distances
  for (disti = 0; disti < dists.length; disti++)
    {
      if (dists[disti]["enabled"] != 1)
	continue;

      fval = parseFloat (document.getElementById ("dist_" + disti).value);
      if (isNaN (fval))
	continue;

      obji = dists[disti]["obj1"];
      if (obji < 0)
	obj1str = dists[disti]["cam1"];
      else
	{
	  if (objs[obji]["enabled"] != 1)
	    continue;
	  obj1str = "Refpt_R" + obji;
	}

      obji = dists[disti]["obj2"];
      if (obji < 0)
	obj2str = dists[disti]["cam2"];
      else
	{
	  if (objs[obji]["enabled"] != 1)
	    continue;
	  obj2str = "Refpt_R" + obji;
	}

      msgadd = {
      "obj":obj1str,
      "obj2":obj2str,
      "d":fval
      };
      msg.push (msgadd);
    }

  mqttmsg = new Paho.MQTT.Message (JSON.stringify (msg, null, 2));
  mqttmsg.destinationName = topicprefix + "/scene/observ";
  mqttclient.send (mqttmsg);

  // Clear previous results
  positionresult_text_new = "";

  sceneresult_text_new = "&nbsp;<br/>Scene observations submitted...";

  scenedetails_text_new = "";

  scenecalc_running = true;

  //console.log ("submit: " + msg);
}

// Clickable link
function
submitScenecalc ()
{
  if (!window.confirm ("Really ready to submit to scene calculation?"))
    return;

  submitScenecalcDo ();
}

// Clickable link
function
stopScenecalc ()
{
  if (!window.confirm ("Really stop scene calculation?"))
    return;

  mqttmsg = new Paho.MQTT.Message ('{"cmd": "stopcalc"}');
  mqttmsg.destinationName = topicprefix + "/scene/cmd";
  mqttclient.send (mqttmsg);

  // Do not overwrite actual result, if that was posted already.
  if (sceneresult_text_new.indexOf ("observations submitted") > 0)
    sceneresult_text_new = "&nbsp;<br/>Scene calculation stopped.";

  scenecalc_running = false;
}

function
escapeHtml (unsafe)
{
  // https://stackoverflow.com/questions/6234773
  // *INDENT-OFF*
  return unsafe
	.replace(/&/g, "&amp;")
	.replace(/</g, "&lt;")
	.replace(/>/g, "&gt;")
	.replace(/"/g, "&quot;")
	.replace(/'/g, "&#039;");
  // *INDENT-ON*
}


// Camera data reception
//---------------------------------------------------------------------------

function
newCam (camname)
{
  var newHtml;
  var camname2;
  var opt;

  newHtml = "<div id=\"cammain_" + camname + "\" class=\"cammain_normal\""
    + " onclick=\"camFullscrToggle(this,'" + camname + "');\">"
    + " <div class=\"camimgdiv\">"
    + "   <img id=\"camimg_orig_" + camname
    + "\" class=\"camimg\" src=\"\" alt=\"\">"
    + "   <img id=\"camimg_colfilt_" + camname
    + "\" class=\"camimg\" src=\"\" alt=\"\">"
    + "   <img id=\"camimg_campix_" + camname
    + "\" class=\"cammarker\" src=\"\" alt=\"\">"
    + "   <img id=\"camimg_calcpix_" + camname
    + "\" class=\"cammarker\" src=\"\" alt=\"\">"
    + "   <img id=\"camimg_obj1_" + camname
    + "\" class=\"cammarker\" src=\"\" alt=\"\">"
    + "   <img id=\"camimg_obj2_" + camname
    + "\" class=\"cammarker\" src=\"\" alt=\"\">"
    + " </div>"
    + " <div class=\"caminfoline\">"
    + "Cam " + camname
    + "  <input type=\"checkbox\" onclick=\"camViewColfilt(this,event,'"
    + camname + "');\" title=\"Show Colfilt image\">"
    + "  <span id=\"caminfoline_" + camname
    + "\"></span>" + " </div>" + "</div>";

  document.getElementById ("camspace").insertAdjacentHTML ("beforeend",
							   newHtml);

  newHtml = "<div id=\"camheightdiv_" + camname + "\"><label"
    + " for=\"camheight_" + camname + "\">Cam " + camname
    + " </label><input id=\"camheight_" + camname
    + "\" name=\"camheight_" + camname + "\">"
    + " <span class=\"vishid\">X</span>" + "</div>";

  document.getElementById ("camheightspace").insertAdjacentHTML ("beforeend",
								 newHtml);

  opt = document.createElement ("option");
  opt.value = "cam_" + camname;
  opt.text = "Cam " + camname;
  document.getElementById ("seldist_o1").add (opt);

  opt = document.createElement ("option");
  opt.value = "cam_" + camname;
  opt.text = "Cam " + camname;
  document.getElementById ("seldist_o2").add (opt);


  cams[camname] = {
	// *INDENT-OFF*
	pixx: -1,
	pixy: -1,
	pixw: 1,
	pixh: 1,
	calcpixx: -1,
	calcpixy: -1,
	fps: 0,
	blinkspersec: 0,
	blinkthres: 0,
	blinkcolorr: 0,
	blinkcolorg: 0,
	blinkcolorb: 0,
	img_orig: "",
	img_orig_reload: 0,
	img_colfilt: "",
	img_colfilt_reload: 0,
	view_whichimg: 0,
	// *INDENT-ON*
  };

  // Make really separate var: .bind() stores pointer to current
  // _environment_ and only evaluates at exec time.
  camname2 = camname;
  cams[camname]["pix_invalid_timer"] =
    window.setTimeout (invalidateCampix.bind (null, camname2), 2000);
  cams[camname]["calcpix_invalid_timer"] =
    window.setTimeout (invalidateCalcpix.bind (null, camname2), 2000);
  // Note: re-armed in handleCamMessage

  cams[camname]["infoline_el"] =
    document.getElementById ("caminfoline_" + camname);

  cams[camname]["img_orig_el"] =
    document.getElementById ("camimg_orig_" + camname);
  cams[camname]["img_orig_el"].style.display = "block";

  cams[camname]["img_colfilt_el"] =
    document.getElementById ("camimg_colfilt_" + camname);
  cams[camname]["img_colfilt_el"].style.display = "none";

  cams[camname]["img_campix_el"] =
    document.getElementById ("camimg_campix_" + camname);
  cams[camname]["img_campix_el"].style.display = "none";
  cams[camname]["img_campix_el"].src = img_campix_url;

  cams[camname]["img_calcpix_el"] =
    document.getElementById ("camimg_calcpix_" + camname);
  cams[camname]["img_calcpix_el"].style.display = "none";
  cams[camname]["img_calcpix_el"].src = img_calcpix_url;

  cams[camname]["img_obj1_el"] =
    document.getElementById ("camimg_obj1_" + camname);
  cams[camname]["img_obj1_el"].style.display = "none";
  cams[camname]["img_obj1_el"].src = img_obj1_url;

  cams[camname]["img_obj2_el"] =
    document.getElementById ("camimg_obj2_" + camname);
  cams[camname]["img_obj2_el"].style.display = "none";
  cams[camname]["img_obj2_el"].src = img_obj2_url;

  cams[camname]["img_orig_precacheimg"] = new Image ();
  cams[camname]["img_colfilt_precacheimg"] = new Image ();

  // Request initial image immediately
  requestImgsLoopName = "cam/" + camname;
  requestImgsLoopForce = true;
  requestImgsLoop ();
}


var whiteSplit = new RegExp ("\\s+");

function
handleCamMessage (camname, subtopic, furthertopics, msgstring)
{
  var camname2;
  var postfix;
  var jmsg;
  var jblinkcolorrgb;
  var imgloc;

  //console.log ("handleCamMessage for " + camname + " sub " + subtopic +
  //           " further " + furthertopics + " msg " + msgstring);

  try
  {
    jmsg = JSON.parse (msgstring);
  }
  catch (error)
  {
    return;
  }
  //console.log("jmsg: %o <--click on it!", jmsg);

  if (!Object.prototype.hasOwnProperty.call (cams, camname))
    newCam (camname);

  switch (subtopic)
    {
    case "pix":
      window.clearTimeout (cams[camname]["pix_invalid_timer"]);

      if (jmsg.hasOwnProperty ("pixx") &&
	  jmsg.hasOwnProperty ("pixy") &&
	  Number.isFinite (jmsg["pixx"]) && Number.isFinite (jmsg["pixy"]))
	{
	  cams[camname]["pixx"] = jmsg["pixx"];
	  cams[camname]["pixy"] = jmsg["pixy"];

	  if (jmsg.hasOwnProperty ("pixw") &&
	      jmsg.hasOwnProperty ("pixh") &&
	      Number.isFinite (jmsg["pixw"]) &&
	      Number.isFinite (jmsg["pixh"]))
	    {
	      cams[camname]["pixw"] = jmsg["pixw"];
	      cams[camname]["pixh"] = jmsg["pixh"];
	    }
	  else
	    {
	      cams[camname]["pixw"] = 1;
	      cams[camname]["pixh"] = 1;
	    }

	  // Restart fix-lost timeout
	  // Make really separate var: .bind() stores pointer to current
	  // _environment_ and only evaluates at exec time.
	  camname2 = camname;
	  cams[camname]["pix_invalid_timer"] =
	    window.setTimeout (invalidateCampix.bind (null, camname2), 2000);
	}
      else if (jmsg.hasOwnProperty ("lost") && jmsg["lost"] == "connection")
	{
	  cams[camname]["pixx"] = -2;
	  cams[camname]["pixy"] = -2;
	}
      else
	{
	  // "lost":"fix" or anything else
	  cams[camname]["pixx"] = -1;
	  cams[camname]["pixy"] = -1;
	}

      break;

    case "calcpix":
      window.clearTimeout (cams[camname]["calcpix_invalid_timer"]);

      if (jmsg.hasOwnProperty ("pixx") && jmsg.hasOwnProperty ("pixy") &&
	  Number.isFinite (jmsg["pixx"]) && Number.isFinite (jmsg["pixy"]))
	{
	  cams[camname]["calcpixx"] = jmsg["pixx"];
	  cams[camname]["calcpixy"] = jmsg["pixy"];
	}
      else
	{
	  // Fix-lost
	  cams[camname]["calcpixx"] = -1;
	  cams[camname]["calcpixy"] = -1;
	}

      // Restart fix-lost timeout
      // Make really separate var: .bind() stores pointer to current
      // _environment_ and only evaluates at exec time.
      camname2 = camname;
      cams[camname]["calcpix_invalid_timer"] =
	window.setTimeout (invalidateCalcpix.bind (null, camname2), 2000);

      break;

    case "status":
      if (jmsg.hasOwnProperty ("fps") && Number.isFinite (jmsg["fps"]))
	cams[camname]["fps"] = jmsg["fps"];

      if (jmsg.hasOwnProperty ("blinkspersec") &&
	  Number.isFinite (jmsg["blinkspersec"]))
	cams[camname]["blinkspersec"] = jmsg["blinkspersec"];

      if (jmsg.hasOwnProperty ("blinkthres") &&
	  Number.isFinite (jmsg["blinkthres"]))
	cams[camname]["blinkthres"] = jmsg["blinkthres"];

      if (jmsg.hasOwnProperty ("blinkcolor") &&
	  jmsg["blinkcolor"].hasOwnProperty ("rgb"))
	{
	  jblinkcolorrgb = jmsg["blinkcolor"]["rgb"];
	  if (jblinkcolorrgb.hasOwnProperty ("r") &&
	      jblinkcolorrgb.hasOwnProperty ("g") &&
	      jblinkcolorrgb.hasOwnProperty ("b") &&
	      Number.isFinite (jblinkcolorrgb["r"]) &&
	      Number.isFinite (jblinkcolorrgb["g"]) &&
	      Number.isFinite (jblinkcolorrgb["b"]))
	    {
	      cams[camname]["blinkcolorr"] = jblinkcolorrgb["r"];
	      cams[camname]["blinkcolorg"] = jblinkcolorrgb["g"];
	      cams[camname]["blinkcolorb"] = jblinkcolorrgb["b"];
	    }
	}

    case "img":
      if (furthertopics.length == 4)
	{
	  if (furthertopics[1] != "jpeg" && furthertopics[1] != "png")
	    break;
	  if (furthertopics[3] != "loc")
	    break;

	  //XXXTODO: specifiable imageprocessor name, subscribe & check

	  if (window.location.protocol == "file:" &&
	      jmsg.hasOwnProperty ("file"))
	    imgloc = jmsg["file"];
	  else if (window.location.protocol != "file:" &&
		   jmsg.hasOwnProperty ("url"))
	    imgloc = jmsg["url"];
	  else
	    break;

	  postfix = "";
	  //if (window.location.protocol != "file:")
	  // this is actually also required for local files...
	  postfix = "?r=" + Math.random ();

	  if (furthertopics[0] == "orig")
	    {
	      cams[camname]["img_orig_reload"] = 1;
	      cams[camname]["img_orig"] = imgloc + postfix;

	      if (cams[camname]["view_whichimg"] == 0)
		cams[camname]["img_orig_precacheimg"].src =
		  cams[camname]["img_orig"];
	    }
	  else if (furthertopics[0] == "colfilt")
	    {
	      cams[camname]["img_colfilt_reload"] = 1;
	      cams[camname]["img_colfilt"] = imgloc + postfix;

	      if (cams[camname]["view_whichimg"] == 1)
		cams[camname]["img_colfilt_precacheimg"].src =
		  cams[camname]["img_colfilt"];
	    }

	  expectingProcIdle = "imgproc/" + furthertopics[2];
	}
      break;
    }

  // Note that we just store the values, since immediate updates
  // would be too heavy -- and not needed.
  // HTML is updated via separate timer loop.

  //console.log ("cams: %o <--click on it for *current* values", cams);
}


// Timer-triggered pixel invalidation
function
invalidateCampix (camname)
{
  window.clearTimeout (cams[camname]["pix_invalid_timer"]);
  if (cams[camname]["pixx"] > -0.5)
    {
      cams[camname]["pixx"] = -1;
      cams[camname]["pixy"] = -1;
    }
  // else: keep either "-1 -1" or "-2 -2" as previously reported.
}

function
invalidateCalcpix (camname)
{
  window.clearTimeout (cams[camname]["calcpix_invalid_timer"]);
  cams[camname]["calcpixx"] = -1;
  cams[camname]["calcpixy"] = -1;
}


// Camera visual settings
//---------------------------------------------------------------------------

// Checkbox triggered switch: orig vs colfilt image
function
camViewColfilt (el, event, camname)
{
  event.stopPropagation ();
  event.stopImmediatePropagation ();
  // Note: does not work when onchange: onclick on parent will still fire.

  if (!(el.checked))
    {
      cams[camname]["view_whichimg"] = 0;
      cams[camname]["img_orig_el"].style.display = "block";
      cams[camname]["img_colfilt_el"].style.display = "none";
    }
  else
    {
      cams[camname]["view_whichimg"] = 1;
      cams[camname]["img_orig_el"].style.display = "none";
      cams[camname]["img_colfilt_el"].style.display = "block";
    }

  // Always try to get updated image immediately (~"Reload button")
  requestImgsLoopName = "cam/" + camname;
  requestImgsLoopForce = true;
  requestImgsLoop ();

  updateAllCamHtml ();
}

// Onclick triggered switch: normal vs fullscreen
function
camFullscrToggle (el, camname)
{
  if (el.className.indexOf ("cammain_fullscreen") >= 0)
    {
      el.className = "cammain_normal";
      img_fullscreen = "";
    }
  else
    {
      el.className = "cammain_fullscreen";
      img_fullscreen = "cam/" + camname;

      // Get updated image immediately; and will now do the fullscreen img only.
      requestImgsLoopName = img_fullscreen;
      requestImgsLoopForce = true;
      requestImgsLoop ();
    }

  updateAllCamHtml ();
}

// Clickable variants
function
camimgPerRow (percent)
{
  var theRules;
  var i;

  // Finding and editing CSS rules:
  // https://stackoverflow.com/questions/730048/how-to-change-remove-css-classes-definitions-at-runtime

  // (note that [0] is because the .css file is loaded first in the html file)
  try
  {
    theRules = document.styleSheets[0].cssRules;
    for (i = 0; i < theRules.length; i++)
      if (theRules[i].selectorText == "div.cammain_normal")
	{
	  theRules[i].style.width = percent;
	}
  }
  catch (error)
  {
    alert ("This may fail when using file:// to access page.\n"
	   + "Set preference privacy.file_unique_origin to false.\n"
	   +
	   "See https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS/Errors/CORSRequestNotHttp\n"
	   +
	   "\nMay also fail when using extension that adds stylesheet; then disable that extension.\n");
  }

  updateAllCamHtml ();
  //console.log ("set to " + percent);
}

function
imgPerRowSelect (el)
{
  camimgPerRow ((Math.floor (100 / el.value)) + "%");
}


// Collector (render) data reception
//---------------------------------------------------------------------------

// This is similar to the camera stuff.

function
newCollect (collectname)
{
  collects[collectname] = {
	// *INDENT-OFF*
	dummy: "",
	// *INDENT-ON*
  };
}

function
newRender (collectname, rendername)
{
  var newHtml;
  var opt;

  newHtml = "<div id=\"collectmain_" + collectname + "_" + rendername
    + "\" class=\"cammain_normal\""
    + " onclick=\"collectFullscrToggle(this,'" + collectname + "','"
    + rendername + "');\">"
    + " <div class=\"camimgdiv\">"
    + "   <img id=\"collectimg_" + collectname + "_" + rendername
    + "\" class=\"camimg\" src=\"\" alt=\"\">"
    + " </div>"
    + " <div class=\"caminfoline\">"
    + "Render " + rendername
    + "  <span id=\"collectinfoline_" + collectname + "_" + rendername
    + "\"></span>" + " </div>" + "</div>";

  document.getElementById ("camspace").insertAdjacentHTML ("beforeend",
							   newHtml);

  collects[collectname]["img_" + rendername] = {
	// *INDENT-OFF*
	url: "",
	reload: 0,
	// *INDENT-ON*
  };

  collects[collectname]["img_" + rendername]["infoline_el"] =
    document.getElementById ("collectinfoline_" + collectname
			     + "_" + rendername);

  collects[collectname]["img_" + rendername]["img_el"] =
    document.getElementById ("collectimg_" + collectname + "_" + rendername);
  collects[collectname]["img_" + rendername]["img_el"].style.display =
    "block";

  collects[collectname]["img_" + rendername]["precacheimg"] = new Image ();
}

function
handleCollectMessage (collectname, subtopic, furthertopics, msgstring)
{
  var rendername;
  var postfix;
  var jmsg;
  var imgloc;

  //console.log ("handleCollectMessage for " + collectname + " sub " + subtopic +
  //           " further " + furthertopics + " msg " + msgstring);

  try
  {
    jmsg = JSON.parse (msgstring);
  }
  catch (error)
  {
    return;
  }
  //console.log("jmsg: %o <--click on it!", jmsg);

  if (!Object.prototype.hasOwnProperty.call (collects, collectname))
    newCollect (collectname);

  switch (subtopic)
    {
    case "img":
      if (furthertopics.length == 4)
	{
	  if (furthertopics[1] != "jpeg" && furthertopics[1] != "png")
	    break;
	  if (furthertopics[3] != "loc")
	    break;

	  //XXXTODO: specifiable imageprocessor name, subscribe & check

	  if (window.location.protocol == "file:" &&
	      jmsg.hasOwnProperty ("file"))
	    imgloc = jmsg["file"];
	  else if (window.location.protocol != "file:" &&
		   jmsg.hasOwnProperty ("url"))
	    imgloc = jmsg["url"];
	  else
	    break;

	  rendername = furthertopics[0];
	  if (!Object.prototype.hasOwnProperty.call (collects[collectname],
						     "img_" + rendername))
	    newRender (collectname, rendername);

	  postfix = "";
	  //if (window.location.protocol != "file:")
	  // this is actually also required for local files...
	  postfix = "?r=" + Math.random ();

	  collects[collectname]["img_" + rendername]["reload"] = 1;
	  collects[collectname]["img_" + rendername]["url"] =
	    imgloc + postfix;

	  collects[collectname]["img_" + rendername]["precacheimg"].src =
	    collects[collectname]["img_" + rendername]["url"];

	  expectingProcIdle = "render/" + furthertopics[2];
	}
      break;
    }

  // Note that we just store the values, since immediate updates
  // would be too heavy -- and not needed.
  // HTML is updated via separate timer loop.

  //console.log ("collects: %o <--click on it for *current* values", collects);
}


// Imageprocessing data reception
//---------------------------------------------------------------------------

function
handleImgprocMessage (imgprocname, subtopic, msgstring)
{
  var jmsg;

  try
  {
    jmsg = JSON.parse (msgstring);
  }
  catch (error)
  {
    return;
  }
  //console.log("jmsg: %o <--click on it!", jmsg);

  switch (subtopic)
    {
    case "status":
      if (jmsg.hasOwnProperty ("state"))
	if (jmsg["state"] == "idle")
	  image_processing_idle ("imgproc/" + imgprocname);
      break;
    }
}


// Renderer data reception
//---------------------------------------------------------------------------

function
handleRenderMessage (rendername, subtopic, msgstring)
{
  var jmsg;

  try
  {
    jmsg = JSON.parse (msgstring);
  }
  catch (error)
  {
    return;
  }
  //console.log("jmsg: %o <--click on it!", jmsg);

  switch (subtopic)
    {
    case "status":
      if (jmsg.hasOwnProperty ("state"))
	if (jmsg["state"] == "idle")
	  image_processing_idle ("render/" + rendername);
      break;
    }
}


// Collector (render) visual settings
//---------------------------------------------------------------------------

// Onclick triggered switch: normal vs fullscreen
function
collectFullscrToggle (el, collectname, rendername)
{
  if (el.className.indexOf ("cammain_fullscreen") >= 0)
    {
      el.className = "cammain_normal";
      img_fullscreen = "";
    }
  else
    {
      el.className = "cammain_fullscreen";
      img_fullscreen = "collect/" + collectname;	// not rendername!

      // Get updated image immediately; and will now do the fullscreen img only.
      requestImgsLoopName = img_fullscreen;
      requestImgsLoopForce = true;
      requestImgsLoop ();
    }

  updateAllCamHtml ();
}


// Reference object positions save/show/delete
//---------------------------------------------------------------------------

function
preventNavigatingAway ()
{
  window.onbeforeunload = function ()
  {
    return
      "Do you really want to close this page? Entered data will be lost!";
  }
}

// Called from mouseclick
function
addObj ()
{
  var newHtml;
  var objnum;
  var objobj;
  var camNames;
  var cami;
  var camname;
  var opt;

  preventNavigatingAway ();

  objobj = {
	// *INDENT-OFF*
	enabled:1,
	// *INDENT-ON*
  };
  objobj["cams"] = {
  };
  objnum = objs.push (objobj) - 1;

  newHtml = "<div id=\"objheightdiv_" + objnum + "\""
    + " class=\"objheightitem\""
    + " onmouseover=\"objMouseOver(" + objnum + ");\""
    + " onmouseout=\"objMouseOut(" + objnum + ");\""
    + "><label"
    + " for=\"objheight_" + objnum + "\">Refpt R" + objnum
    + " </label><input id=\"objheight_" + objnum
    + "\" name=\"objheight_" + objnum + "\">"
    + " <a href=\"javascript:delObj(" + objnum + ");\">X</a>" + "</div>";

  document.getElementById ("objheightspace").insertAdjacentHTML ("beforeend",
								 newHtml);

  // Add to distance selectboxes

  opt = document.createElement ("option");
  opt.value = "obj_" + objnum;
  opt.text = "Refpt R" + objnum;
  opt.onmouseover = function (ev)
  {
    distSelectMouseOver (objnum, -1);
    ev.stopPropagation ();
  };
  document.getElementById ("seldist_o1").add (opt);

  opt = document.createElement ("option");
  opt.value = "obj_" + objnum;
  opt.text = "Refpt R" + objnum;
  opt.onmouseover = function (ev)
  {
    distSelectMouseOver (-1, objnum);
    ev.stopPropagation ();
  };
  document.getElementById ("seldist_o2").add (opt);

  // Store current camera pixels
  camNames = Object.getOwnPropertyNames (cams);
  for (cami = 0; cami < camNames.length; cami++)
    {
      camname = camNames[cami];
      if (cams[camname]["pixx"] >= 0 && cams[camname]["pixy"] >= 0)
	{
	  objobj["cams"][camname] = {
	  };
	  objobj["cams"][camname]["pixx"] = cams[camname]["pixx"];
	  objobj["cams"][camname]["pixy"] = cams[camname]["pixy"];
	  objobj["cams"][camname]["pixw"] = cams[camname]["pixw"];
	  objobj["cams"][camname]["pixh"] = cams[camname]["pixh"];
	}
    }

  //console.log ("objs: %o <--click on it for *current* values", objs);
}

// Onmouseover/out triggers
function
objMouseOver (objnum)
{
  var redraw;

  //console.log ("mouseover %d", objnum);

  // Note: we are also called when moving from inputbox to label, and
  // from label to div's surrounding whitespace. Only redraw when changed.
  redraw = 1;
  if (obj1_num == objnum && obj2_num < 0)
    redraw = 0;

  obj1_num = objnum;
  obj2_num = -1;

  if (redraw)
    updateAllCamHtml ();
}

function
objMouseOut (objnum)
{
  //console.log ("mouseout %d", objnum);

  obj1_num = -1;
  obj2_num = -1;

  // Do not trigger immediate redraw, either on mouseover to next item,
  // else in normal 0.5sec loop.
}

// Called from mouseclick
function
delObj (objnum)
{
  var obj_el;
  var sel_el;
  var sel_options;
  var i;

  if (!window.confirm ("Really delete Refpoint R" + objnum
		       + " and all related distances?"))
    return;

  objs[objnum]["enabled"] = 0;

  obj_el = document.getElementById ("objheightdiv_" + objnum);
  obj_el.parentNode.removeChild (obj_el);

  // Usually pointer will be on deleted object div
  obj1_num = -1;
  obj2_num = -1;


  // Also delete from distance selectboxes

  sel_el = document.getElementById ("seldist_o1");
  sel_options = sel_el.options;
  for (i = 0; i < sel_options.length; i++)
    if (sel_options[i].value == "obj_" + objnum)
      {
	sel_el.remove (i);
	break;			// Only one; and .length possibly changed
      }

  sel_el = document.getElementById ("seldist_o2");
  sel_options = sel_el.options;
  for (i = 0; i < sel_options.length; i++)
    if (sel_options[i].value == "obj_" + objnum)
      {
	sel_el.remove (i);
	break;			// Only one; and .length possibly changed
      }


  // Also remove all related distances

  for (i = 0; i < dists.length; i++)
    if (dists[i]["enabled"] == 1 &&
	(dists[i]["obj1"] == objnum || dists[i]["obj2"] == objnum))
      {
	dists[i]["enabled"] = 0;

	obj_el = document.getElementById ("distdiv_" + i);
	obj_el.parentNode.removeChild (obj_el);
      }

}


// Distances save/show/delete
//---------------------------------------------------------------------------

// Called from mouseclick
function
addDist ()
{
  var newHtml;
  var distnum;
  var distobj;
  var sel_val;

  preventNavigatingAway ();

  distobj = {
	// *INDENT-OFF*
	enabled:1,
	// *INDENT-ON*
  };

  sel_val = document.getElementById ("seldist_o1").value;
  if (sel_val.startsWith ("cam_"))
    {
      distobj["obj1"] = -1;	// =look at cam
      distobj["cam1"] = sel_val.replace ("cam_", "");
      distobj["text1"] = "" + distobj["cam1"];
    }
  else if (sel_val.startsWith ("obj_"))
    {
      distobj["obj1"] = Number (sel_val.replace ("obj_", ""));
      distobj["text1"] = "R" + distobj["obj1"];
    }
  else
    // Something is wrong
    return;

  sel_val = document.getElementById ("seldist_o2").value;
  if (sel_val.startsWith ("cam_"))
    {
      distobj["obj2"] = -1;	// =look at cam
      distobj["cam2"] = sel_val.replace ("cam_", "");
      distobj["text2"] = "" + distobj["cam2"];
    }
  else if (sel_val.startsWith ("obj_"))
    {
      distobj["obj2"] = Number (sel_val.replace ("obj_", ""));
      distobj["text2"] = "R" + distobj["obj2"];
    }
  else
    // Something is wrong
    return;

  // Check if same
  if (distobj["obj1"] < 0 && distobj["obj2"] < 0)
    {
      if (distobj["cam1"] == distobj["cam2"])
	return;
    }
  else if (distobj["obj1"] == distobj["obj2"])
    return;

  distnum = dists.push (distobj) - 1;

  newHtml = "<div id=\"distdiv_" + distnum + "\""
    + " class=\"distitem\""
    + " onmouseover=\"distMouseOver(" + distnum + ");\""
    + " onmouseout=\"distMouseOut(" + distnum + ");\""
    + "><label"
    + " for=\"dist_" + distnum + "\">" + distobj["text1"]
    + " - " + distobj["text2"]
    + " </label><input id=\"dist_" + distnum
    + "\" name=\"dist_" + distnum + "\">"
    + " <a href=\"javascript:delDist(" + distnum + ");\">X</a>" + "</div>";

  document.getElementById ("distspace").insertAdjacentHTML ("beforeend",
							    newHtml);

  //console.log ("dists: %o <--click on it for *current* values", dists);
}

// Onmouseover/out triggers
function
distMouseOver (distnum)
{
  var obj1;
  var obj2;
  var redraw;

  obj1 = dists[distnum]["obj1"];	// -1 if cam
  obj2 = dists[distnum]["obj2"];

  //console.log ("mouseover %d", distnum);

  // Note: we are also called when moving from inputbox to label, and
  // from label to div's surrounding whitespace. Only redraw when changed.
  redraw = 1;
  if (obj1_num == obj1 && obj2_num == obj2)
    redraw = 0;

  obj1_num = obj1;
  obj2_num = obj2;

  if (redraw)
    updateAllCamHtml ();
}

function
distMouseOut (distnum)
{
  //console.log ("mouseout %d", distnum);

  obj1_num = -1;
  obj2_num = -1;

  // Do not trigger immediate redraw, either on mouseover to next item,
  // else in normal 0.5sec loop.
}

// Called from mouseclick
function
delDist (distnum)
{
  var obj_el;

  if (!window.confirm ("Really delete distance " + dists[distnum]["text1"]
		       + " - " + dists[distnum]["text2"] + " ?"))
    return;

  dists[distnum]["enabled"] = 0;

  obj_el = document.getElementById ("distdiv_" + distnum);
  obj_el.parentNode.removeChild (obj_el);

  // Usually pointer will be on deleted object div
  obj1_num = -1;
  obj2_num = -1;
}

function
distSelectMouseOver (obj1_override, obj2_override)
{
  var obj1;
  var obj2;
  var sel_val;
  var redraw;

  obj1 = -1;
  obj2 = -1;

  if (obj1_override >= 0)
    obj1 = obj1_override;
  else
    {
      sel_val = document.getElementById ("seldist_o1").value;
      if (sel_val.startsWith ("obj_"))
	obj1 = Number (sel_val.replace ("obj_", ""));
    }

  if (obj2_override >= 0)
    obj2 = obj2_override;
  else
    {
      sel_val = document.getElementById ("seldist_o2").value;
      if (sel_val.startsWith ("obj_"))
	obj2 = Number (sel_val.replace ("obj_", ""));
    }

  redraw = 1;
  if (obj1_num == obj1 && obj2_num == obj2)
    redraw = 0;

  if (0)
    console.log ("dsmouseover ovr %d %d obj %d %d re %d", obj1_override,
		 obj2_override, obj1, obj2, redraw);

  obj1_num = obj1;
  obj2_num = obj2;

  if (redraw)
    updateAllCamHtml ();
}

function
distSelectMouseOut ()
{
  obj1_num = -1;
  obj2_num = -1;

  // Do not trigger immediate redraw, either on mouseover to next item,
  // else in normal 0.5sec loop.
}


// Visual rendering
//---------------------------------------------------------------------------

function
updateCamHtml (camname)
{
  var infoline;
  var imgel;
  var obj_pixx;
  var obj_pixy;

  infoline = "";

  if (cams[camname]["pixx"] > -0.5)
    {
      infoline +=
	"<span class=\"nowrap\" style=\"background: #aaffaa\">Fix "
	+ cams[camname]["pixx"].toFixed (1) + " "
	+ cams[camname]["pixy"].toFixed (1) + " "
	+ cams[camname]["pixw"].toFixed (0) + "x"
	+ cams[camname]["pixh"].toFixed (0) + "</span> ";
    }
  else if (cams[camname]["pixx"] < -1.5)	// -2 mostly
    {
      infoline +=
	"<span class=\"nowrap\" style=\"color: white; background: #d00000\">"
	+ "Connection Lost</span> ";
    }
  else				// -1 mostly
    {
      infoline +=
	"<span class=\"nowrap\" style=\"background: #ffbbbb\">"
	+ "Fix Lost</span> ";
    }

  infoline += "<span class=\"nowrap\">Calc "
    + cams[camname]["calcpixx"].toFixed (1) + " "
    + cams[camname]["calcpixy"].toFixed (1) + "</span> "
    // Note that we set calcpix to -1 on timeout, but that is a totally
    // valid value: calculation just considers frame boundaries an occlusion.
    + "<span class=\"nowrap\">Fps "
    + cams[camname]["fps"].toFixed (0)
    + "</span> "
    + "<span class=\"nowrap\">Bl/s "
    + cams[camname]["blinkspersec"].toFixed (1)
    + "</span> "
    + "<span class=\"nowrap\">Thr "
    + cams[camname]["blinkthres"].toFixed (0)
    + "</span> "
    + "<span class=\"nowrap\">RGB "
    + cams[camname]["blinkcolorr"].toFixed (0) + " "
    + cams[camname]["blinkcolorg"].toFixed (0) + " "
    + cams[camname]["blinkcolorb"].toFixed (0) + "</span> ";

  cams[camname]["infoline_el"].innerHTML = infoline;


  // Take over any new image, should already be in cache.
  // Note: .style.display = block/none is in camViewColfilt()

  imgel = "img_orig_el";
  if (cams[camname].view_whichimg == 0)
    {
      imgel = "img_orig_el";

      if (cams[camname]["img_orig_reload"] == 1)
	{
	  cams[camname]["img_orig_el"].src = cams[camname]["img_orig"];
	  cams[camname]["img_orig_reload"] = 0;
	}
    }
  else if (cams[camname].view_whichimg == 1)
    {
      imgel = "img_colfilt_el";

      if (cams[camname]["img_colfilt_reload"] == 1)
	{
	  cams[camname]["img_colfilt_el"].src = cams[camname]["img_colfilt"];
	  cams[camname]["img_colfilt_reload"] = 0;
	}
    }


  // Position markers
  // Use imgel for automatic swap according to displayed image

  // Camera detected blink pixel

  if (cams[camname]["pixx"] >= 0 &&
      cams[camname]["pixy"] >= 0 &&
      img_campix.height > 2 &&
      cams[camname][imgel].height > 2 &&
      cams[camname]["pixx"] < cams[camname][imgel].naturalWidth &&
      cams[camname]["pixy"] < cams[camname][imgel].naturalHeight)
    {
      cams[camname]["img_campix_el"].style.left =
	cams[camname][imgel].width * cams[camname]["pixx"] /
	cams[camname][imgel].naturalWidth - img_campix.width / 2 + "px";

      cams[camname]["img_campix_el"].style.top =
	cams[camname][imgel].height * cams[camname]["pixy"] /
	cams[camname][imgel].naturalHeight - img_campix.height / 2 + "px";

      cams[camname]["img_campix_el"].style.display = "block";
    }
  else
    {
      cams[camname]["img_campix_el"].style.display = "none";
    }

  // Back-calculated blink pixel
  // Note that backcalc pix may very well go out-of-image when object is
  // still tracked by other cams. We don't display the marker then.

  if (cams[camname]["calcpixx"] >= 0 &&
      cams[camname]["calcpixy"] >= 0 &&
      img_calcpix.height > 2 &&
      cams[camname][imgel].height > 2 &&
      cams[camname]["calcpixx"] < cams[camname][imgel].naturalWidth &&
      cams[camname]["calcpixy"] < cams[camname][imgel].naturalHeight)
    {
      cams[camname]["img_calcpix_el"].style.left =
	cams[camname][imgel].width * cams[camname]["calcpixx"] /
	cams[camname][imgel].naturalWidth - img_calcpix.width / 2 + "px";

      cams[camname]["img_calcpix_el"].style.top =
	cams[camname][imgel].height * cams[camname]["calcpixy"] /
	cams[camname][imgel].naturalHeight - img_calcpix.height / 2 + "px";

      cams[camname]["img_calcpix_el"].style.display = "block";
    }
  else
    {
      cams[camname]["img_calcpix_el"].style.display = "none";
    }

  // Saved-object 1

  obj_pixx = -1;
  obj_pixy = -1;
  if (obj1_num >= 0 && obj1_num < objs.length)
    if (objs[obj1_num]["enabled"] == 1 &&
	Object.prototype.hasOwnProperty.call
	(objs[obj1_num]["cams"], camname))
      {
	obj_pixx = objs[obj1_num]["cams"][camname]["pixx"];
	obj_pixy = objs[obj1_num]["cams"][camname]["pixy"];
      }
  if (obj_pixx >= 0 &&
      obj_pixy >= 0 &&
      img_obj1.height > 2 &&
      cams[camname][imgel].height > 2 &&
      obj_pixx < cams[camname][imgel].naturalWidth &&
      obj_pixy < cams[camname][imgel].naturalHeight)
    {
      cams[camname]["img_obj1_el"].style.left =
	cams[camname][imgel].width * obj_pixx /
	cams[camname][imgel].naturalWidth - img_obj1.width / 2 + "px";

      cams[camname]["img_obj1_el"].style.top =
	cams[camname][imgel].height * obj_pixy /
	cams[camname][imgel].naturalHeight - img_obj1.height / 2 + "px";

      cams[camname]["img_obj1_el"].style.display = "block";
    }
  else
    {
      cams[camname]["img_obj1_el"].style.display = "none";
    }

  // Saved-object 2 (for 1-to-2 distance entries)

  obj_pixx = -1;
  obj_pixy = -1;
  if (obj2_num >= 0 && obj2_num < objs.length)
    if (objs[obj2_num]["enabled"] == 1 &&
	Object.prototype.hasOwnProperty.call
	(objs[obj2_num]["cams"], camname))
      {
	obj_pixx = objs[obj2_num]["cams"][camname]["pixx"];
	obj_pixy = objs[obj2_num]["cams"][camname]["pixy"];
      }
  if (obj_pixx >= 0 &&
      obj_pixy >= 0 &&
      img_obj2.height > 2 &&
      cams[camname][imgel].height > 2 &&
      obj_pixx < cams[camname][imgel].naturalWidth &&
      obj_pixy < cams[camname][imgel].naturalHeight)
    {
      cams[camname]["img_obj2_el"].style.left =
	cams[camname][imgel].width * obj_pixx /
	cams[camname][imgel].naturalWidth - img_obj2.width / 2 + "px";

      cams[camname]["img_obj2_el"].style.top =
	cams[camname][imgel].height * obj_pixy /
	cams[camname][imgel].naturalHeight - img_obj2.height / 2 + "px";

      cams[camname]["img_obj2_el"].style.display = "block";
    }
  else
    {
      cams[camname]["img_obj2_el"].style.display = "none";
    }

}

function
updateRenderHtml (collectname, rendername)
{
  // Take over any new image, should already be in cache.
  // Note: .style.display = block/none is in camViewColfilt()

  // This is called for *all* properties;
  // filter only rendernames by lookung at "img_" prefix
  if (rendername.indexOf ("img_") != 0)
    return;

  if (collects[collectname][rendername]["reload"] == 1)
    {
      collects[collectname][rendername]["img_el"].src =
	collects[collectname][rendername]["url"];
      collects[collectname][rendername]["reload"] = 0;
    }
}

function
updateCollectHtml (collectname)
{
  var renderNames;
  var i;

  renderNames = Object.getOwnPropertyNames (collects[collectname]);
  for (i = 0; i < renderNames.length; i++)
    updateRenderHtml (collectname, renderNames[i]);
}


function
setPositionText (newText)
{
  position_text_new = newText;

  // Postponed update every 0.5 sec: firefox 20% CPU
  //return;

  // Immediate update: firefox 22% CPU
  if (position_text_new != position_text)
    {
      position_text = position_text_new;
      document.getElementById ("statusline").innerHTML = position_text;
    }
}


var updateAllCamHtmlTimer = window.setTimeout (updateAllCamHtml, 500);

function
updateAllCamHtml ()
{
  var camNames;
  var collectNames;

  // We may be called out-of-sync; use restartable timer instead of interval.
  window.clearTimeout (updateAllCamHtmlTimer);

  camNames = Object.getOwnPropertyNames (cams);
  camNames.forEach (updateCamHtml);

  collectNames = Object.getOwnPropertyNames (collects);
  collectNames.forEach (updateCollectHtml);

  if (position_text_new != position_text)
    {
      position_text = position_text_new;
      document.getElementById ("statusline").innerHTML = position_text;
    }
  if (positionresult_text_new != positionresult_text)
    {
      positionresult_text = positionresult_text_new;
      document.getElementById ("positionresultdiv").innerHTML =
	positionresult_text;
    }
  if (sceneresult_text_new != sceneresult_text)
    {
      sceneresult_text = sceneresult_text_new;
      document.getElementById ("sceneresultdiv").innerHTML = sceneresult_text;
    }
  if (scenedetails_text_new != scenedetails_text)
    {
      scenedetails_text = scenedetails_text_new;
      document.getElementById ("scenedetailsdiv").innerHTML =
	scenedetails_text;
    }

  updateAllCamHtmlTimer = window.setTimeout (updateAllCamHtml, 500);
}


function
requestImgsLoop ()
{
  var camcollectNames;
  var lastIdx;
  var newIdx;
  var whichimg;
  var message;
  var timerlen;

  // We may be called out-of-sync by user action; use restartable timer
  // instead of interval.
  window.clearTimeout (requestImgsLoopTimer);

  if (lurkermode)
    return;

  // Handle fullscreen as continuously-forced (unless another is forced)
  if (!requestImgsLoopForce && img_fullscreen.length > 0)
    {
      requestImgsLoopName = img_fullscreen;
      requestImgsLoopForce = true;
    }

  // Find next (or forced) cam/collectname
  camcollectNames = Object.getOwnPropertyNames (cams).map (function (el)
							   {
							   return "cam/" + el;
							   }
  ).concat (Object.getOwnPropertyNames (collects).map (function (el)
						       {
						       return "collect/" + el;
						       }
	    ));

  lastIdx = camcollectNames.indexOf (requestImgsLoopName);
  newIdx = 0;
  if (lastIdx >= 0)
    {
      if (!requestImgsLoopForce)
	{
	  // Go to next image
	  newIdx = lastIdx + 1;
	  if (newIdx >= camcollectNames.length)
	    newIdx = 0;
	}
      else
	{
	  // Forced name (that is also correct)
	  newIdx = lastIdx;
	}
    }
  //else: not found, use 0.

  // Only run if there is actually a cam known
  if (newIdx < camcollectNames.length)
    {
      requestImgsLoopName = camcollectNames[newIdx];

      if (requestImgsLoopName.indexOf ("cam/") == 0)
	{
	  whichimg = "orig";
	  if (cams[requestImgsLoopName.slice (4)].view_whichimg == 1)
	    whichimg = "colfilt";

	  message =
	    new Paho.MQTT.Message ('{"cmd": "postimg", "img": "'
				   + whichimg + '"}');
	}
      else if (requestImgsLoopName.indexOf ("collect/") == 0)
	{
	  message = new Paho.MQTT.Message ('{"cmd": "postdata"}');
	}

      message.destinationName =
	topicprefix + "/" + requestImgsLoopName + "/cmd";
      mqttclient.send (message);
    }

  requestImgsLoopForce = false;

  // This is timeout for non-responding images;
  // normally image_processing_idle() will do quick rescheduling.
  timerlen = 5000;

  requestImgsLoopTimer = window.setTimeout (requestImgsLoop, timerlen);
}

function
image_processing_idle (procname)
{
  var timerlen;

  if (lurkermode)
    return;

  //console.log("idle for " + procname + "  expecting " + expectingProcIdle);
  if (procname != expectingProcIdle)
    return;

  // Processing is now idle; reschedule timer to get new image quickly.

  window.clearTimeout (requestImgsLoopTimer);

  if (scenecalc_running && img_fullscreen.length <= 0)
    {
      // Keep CPU available for scenecalc.
      timerlen = 4000;
    }
  else if (img_fullscreen.length > 0)
    {
      // Single image, only displayed every 500ms.
      // Processor already had 100ms idle.
      timerlen = 350;
    }
  else
    {
      // Multiple images visible, can usefully have faster updates.
      timerlen = 200;
    }

  requestImgsLoopTimer = window.setTimeout (requestImgsLoop, timerlen);
}

requestImgsLoopTimer = window.setTimeout (requestImgsLoop, 2000);


// Clickable
function
toggleRightcol (button_el)
{
  var sty;

  sty = document.getElementById ("rightcolumn").style;

  if (sty.display == "block")
    {
      sty.display = "none";
      button_el.innerHTML = "&lt;";
    }
  else
    {
      sty.display = "block";
      button_el.innerHTML = "&gt;";
    }

  updateAllCamHtml ();
}
